# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import cdflib
import numpy as np
import datetime


class Loader:
    def __init__(self, file_path, file_type='cdf', load_data=True):
        self.file_path = file_path
        self.file_type = file_type
        self.variables = {}
        self.metadata = {}
        self.done = False
        if load_data:
            self.load()

    def load(self):
        if self.file_type == 'cdf' or self.file_type == 'hres-cdf':
            self.load_omni_cdf()

    def load_omni_cdf(self):
        f_cdf = cdflib.CDF(self.file_path)
        f_info = f_cdf.cdf_info()
        variables = {}
        self.metadata['var_attrs'] = {}
        for var_name, var_name_cdf in cdf_variable_name_dict.items():
            var = f_cdf.varget(var_name_cdf)
            var_attr = f_cdf.varattsget(var_name_cdf)
            fillval = var_attr['FILLVAL']
            var = np.where(var == fillval, np.nan, var)
            variables[var_name] = np.reshape(var, (var.size, 1))
            self.metadata['var_attrs'].update(var_name=f_cdf.varattsget(var_name_cdf))

        dts_str = cdflib.cdfepoch.encode(variables['EPOCH'].flatten())
        dts = np.empty_like(variables['EPOCH'], dtype=datetime.datetime)
        for ind, dt_str in enumerate(dts_str):
            dts[ind, 0] = datetime.datetime.strptime(dt_str + '000', '%Y-%m-%dT%H:%M:%S.%f')
        variables['DATETIME'] = dts
        variables['B_x_GSM'] = variables['B_x_GSE']
        variables['B_T_GSM'] = np.sqrt(variables['B_z_GSM']**2 + variables['B_y_GSM']**2)
        variables['B_TOTAL'] = np.sqrt(variables['B_z_GSM']**2 + variables['B_y_GSM']**2 + variables['B_x_GSM']**2)

        self.variables = variables
        self.metadata.update(f_cdf.globalattsget(expand=False))
        self.done = True


cdf_variable_name_dict = {
    'EPOCH':    'Epoch',
    'YEAR':     'YR',
    'DAY':      'Day',
    'HOUR':     'HR',
    'MIN':      'Minute',
    'SC_ID_IMF':    'IMF',
    'SC_ID_PLS':    'PLS',
    'IMF_PTS':   'IMF_PTS',
    'PLS_PTS':   'PLS_PTS',
    'PCT_INTERP': 'percent_interp',
    'Timeshift':      'Timeshift',
    'Timeshift_RMS':  'RMS_Timeshift',
    'B_x_GSE':      'BX_GSE',
    'B_y_GSE':      'BY_GSE',
    'B_z_GSE':      'BZ_GSE',
    'B_y_GSM':      'BY_GSM',
    'B_z_GSM':      'BZ_GSM',
    'v_sw':         'flow_speed',
    'v_x':          'Vx',
    'v_y':          'Vy',
    'v_z':          'Vz',
    'n_p':          'proton_density',
    'T':            'T',
    'p_dyn':        'Pressure',
    'E':            'E',
    'beta':         'Beta',
    'Ma_A':         'Mach_num',
    'Ma_MSP':       'Mgs_mach_num',
    'BSN_x':        'BSN_x',
    'BSN_y':        'BSN_y',
    'BSN_z':        'BSN_z',
}


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import pathlib
import h5py
import numpy as np

import geospacelab.toolbox.utilities.pylogging as mylog

# define the default variable name dictionary
default_variable_name_dict = {
    'n_e': 'NE',
    'v_i_H': 'HOR_ION_V',
    'v_i_V': 'VERT_ION_V',
    'B_D': 'BD',
    'B_F': 'B_FORWARD',
    'B_P': 'B_PERP',
    'd_B_D': 'DIFF_BD',
    'd_B_F': 'DIFF_B_FOR',
    'd_B_P': 'DIFF_B_PERP',
    'SC_GEO_LAT': 'GDLAT',
    'SC_GEO_LON': 'GLON',
    'SC_GEO_ALT': 'GDALT',
    'SC_MAG_LAT': 'MLAT',
    'SC_MAG_LON': 'MLONG',
    'SC_MAG_MLT': 'MLT',
}


class Loader(object):
    """
    :param file_path: the full path of the data file
    :type file_path: pathlib.Path or str
    :param file_type: the type of the file, [cdf]
    :param variable_name_dict: the dictionary for mapping the variable names from the cdf files to the dataset
    :type variable_name_dict: dict
    :param direct_load: call the method :meth:`~.LoadModel.load_data` directly or not
    :type direct_load: bool
    """
    def __init__(self, file_path, file_type='hdf5', variable_name_dict=None, direct_load=True, **kwargs):

        self.file_path = pathlib.Path(file_path)
        self.file_ext = file_type
        self.variables = {}
        self.metadata = {}

        if variable_name_dict is None:
            variable_name_dict = default_variable_name_dict
        self.variable_name_dict = variable_name_dict

        if direct_load:
            self.load_data()

    def load_data(self, **kwargs):
        if self.file_ext == 'hdf5':
            self.load_hdf5_data()

    def load_hdf5_data(self):
        """
        load the data from the cdf file
        :return:
        """
        with h5py.File(self.file_path, 'r') as fh5:
            # load metadata
            metadata = {}


            # load data
            data = fh5['Data']['Table Layout'][:]

            data = list(zip(*tuple(data)))

            data_parameters = list(zip(*tuple(fh5['Metadata']['Data Parameters'][:])))
            var_names_h5 = [vn.decode('UTF-8') for vn in data_parameters[0]]
            nvar_h5 = len(var_names_h5)

            vars_h5 = {}
            nrow = len(data[0])
            for ip in range(nvar_h5):
                vars_h5[var_names_h5[ip]] = np.array(data[ip])

            for var_name, var_name_h5 in self.variable_name_dict.items():
                try:
                    self.variables[var_name] = vars_h5[var_name_h5].reshape((nrow, 1))
                except KeyError:
                    print(f"{var_name} is None!")
                    self.variables[var_name] = None

            # add datetime
            dts = [
                datetime.datetime(yy, mm, dd, HH, MM, SS)
                for yy, mm, dd, HH, MM, SS in zip(
                    vars_h5['YEAR'], vars_h5['MONTH'], vars_h5['DAY'],
                    vars_h5['HOUR'], vars_h5['MIN'], vars_h5['SEC']
                )
            ]
            self.variables['SC_DATETIME'] = np.array(dts).reshape((nrow, 1))

            self.metadata = metadata


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import pathlib
import re

from geospacelab import preferences as prf

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.datahub.sources.madrigal.isr.eiscat as eiscat
import geospacelab.toolbox.utilities.pylogging as mylog


default_variable_names = [
    'DATETIME', 'DATETIME_1', 'DATETIME_2',
    'MAGIC_CONSTANT', 'r_SCangle', 'r_m0_1', 'r_m0_2'
    'AZ', 'EL', 'P_Tx', 'HEIGHT', 'RANGE',
    'n_e', 'T_i', 'T_e', 'nu_i', 'v_i_los', 'comp_mix', 'comp_O_p',
    'n_e_err', 'T_i_err', 'T_e_err', 'nu_i_err', 'v_i_los_err', 'comp_mix_err', 'comp_O_p_err',
    'STATUS', 'RESIDUAL'
]


class Loader:
    def __init__(self, file_path, file_type="eiscat-hdf5"):
        self.variables = {}
        self.metadata = {}
        self.file_path = file_path
        self.file_type = file_type

        self.load_data()

    def load_data(self):
        if self.file_type == 'eiscat-hdf5':
            self.load_eiscat_hdf5()
        elif self.file_type == 'eiscat-mat':
            self.load_eiscat_mat()
        elif self.file_type == 'madrigal-hdf5':
            self.load_madrigal_hdf5()
        else:
            raise ValueError

    def load_eiscat_hdf5(self):
        """
        Load EISCAT hdf5 file.
        :param file_paths: a list of the file fullpaths for the hdf5 files
        :return:
            load_obj: instance of Loader
                .variables: queried variables
                .metadata: metadata
        """

        var_name_dict = {
            'MAGIC_CONSTANT': 'Magic_const',
            'r_SCangle': 'SCangle',
            'r_m0_2': 'm02',
            'r_m0_1': 'm01',
            'AZ': 'az',
            'EL': 'el',
            'P_Tx': 'Pt',
            'T_SYS_1': 'Tsys1',
            'T_SYS_2': 'Tsys2',
            'HEIGHT': 'h',
            'RANGE': 'range',
            'n_e': 'Ne',
            'T_i': 'Ti',
            'T_r': 'Tr',
            'nu_i': 'Collf',
            'v_i_los': 'Vi',
            'comp_mix': 'pm',
            'comp_O_p': 'po+',
            'n_e_var': 'var_Ne',
            'T_i_var': 'var_Ti',
            'T_r_var': 'var_Tr',
            'nu_i_var': 'var_Collf',
            'v_i_los_var': 'var_Vi',
            'comp_mix_var': 'var_pm',
            'comp_O_p_var': 'var_po+',
            'STATUS': 'status',
            'RESIDUAL': 'res1'
        }

        site_info = {
            'T': 'UHF',
            'V': 'VHF',
            'K': 'KIR',
            'S': 'SOD',
            'L': 'ESR'
        }

        vars = {}
        metadata = {}

        with h5py.File(self.file_path, 'r') as fh5:
            var_info_list = eiscat.list_eiscat_hdf5_variables(fh5)
            h5_data = fh5['data']
            h5_metadata = fh5['metadata']
            ind_nrec = var_info_list['name'].index('nrec')
            nrec_group = var_info_list['group'][ind_nrec]
            nrec_group_ind = var_info_list['index'][ind_nrec]
            nrec = h5_data[nrec_group][nrec_group_ind]
            num_row = h5_data['utime'][0].shape[0]
            for var_name, var_name_h5 in var_name_dict.items():
                try:
                    ind_v = var_info_list['name'].index(var_name_h5)
                except ValueError:
                    mylog.StreamLogger.warning(f"'{var_name_h5}' is not in the hdf5 file!")
                    vars[var_name] = None
                    continue
                var_group = var_info_list['group'][ind_v]
                var_ind = var_info_list['index'][ind_v]
                var = h5_data[var_group][var_ind]
                if var_group == 'par0d':
                    vars[var_name] = var[0]
                elif var_group == 'par1d':
                    var = var.reshape(num_row, 1)
                    vars[var_name] = var
                elif var_group == 'par2d':
                    if nrec_group == 'par0d':
                        num_col = int(var.shape[0] / num_row)
                        if num_col != nrec[0]:
                            print("Note: the number of range gates doesn't match nrec!")
                        var = var.reshape(num_row, num_col)
                    elif nrec_group == 'par1d':
                        num_gates = int(np.max(nrec))
                        var_array = np.empty((num_row, num_gates))
                        var_array[:, :] = np.nan
                        rec_ind_1 = 0
                        for i in range(num_row):
                            rec_ind_2 = int(rec_ind_1 + nrec[i])
                            var_array[i, :int(nrec[i])] = var[rec_ind_1:rec_ind_2]
                            rec_ind_1 = rec_ind_2
                        var = var_array
                    vars[var_name] = var

            # unix time to datetime
            utime1 = h5_data['utime'][0]
            dt1 = dttool.convert_unix_time_to_datetime_cftime(utime1)
            var = dt1.reshape(num_row, 1)
            var_name = 'DATETIME_1'
            vars[var_name] = var

            utime2 = h5_data['utime'][1]
            dt2 = dttool.convert_unix_time_to_datetime_cftime(utime2)
            var = dt2.reshape(num_row, 1)
            var_name = 'DATETIME_2'
            vars[var_name] = var

            vars['P_Tx'] = vars['P_Tx'] / 1e3   # in kW

            metadata['r_XMITloc'] = [h5_data['par0d'][2][0], h5_data['par0d'][3][0], h5_data['par0d'][4][0]]
            metadata['r_RECloc'] = [h5_data['par0d'][5][0], h5_data['par0d'][6][0], h5_data['par0d'][7][0]]
            metadata['site_name'] = site_info[h5_metadata['names'][1][1].decode('UTF-8').strip()]
            metadata['pulse_code'] = h5_metadata['names'][0][1].decode('UTF-8').strip()
            metadata['antenna'] = h5_metadata['names'][2][1].decode('UTF-8').strip()
            metadata['GUISDAP_version'] = h5_metadata['software']['GUISDAP_ver'][0, 0].decode('UTF-8').strip()
            metadata['rawdata_path'] = ''
            metadata['scan_mode'] = ''
            metadata['affiliation'] = ''
            try:
                title = h5_metadata['schemes']['DataCite']['Title'][0][0].decode('UTF-8').strip()
                rc = re.compile(r'\d{4}-\d{2}-\d{2}_([\S]+)@')
                exp_id = rc.findall(title)[0]
                metadata['modulation'] = exp_id.split('_')[-1]
                metadata['pulse_code'] = exp_id.replace('_'+metadata['modulation'], '')
            except KeyError:
                mylog.StreamLogger.warning("'Title is not listed in the metadata!'")

            try:
                # 'gfd' not in list before 2001?
                metadata['rawdata_path'] = h5_metadata['software']['gfd']['data_path'][0, 0].decode('UTF-8').strip()
                metadata['scan_mode'] = metadata['rawdata_path'].split('_')[1]
                metadata['affiliation'] = metadata['rawdata_path'].split('@')[0].split('_')[-1]
            except KeyError:
                mylog.StreamLogger.warning(
                    "'gfd' is not listed in the metadata! Affect 'rawdata_path', 'scan_mode', and 'affiliation'."
                )

        vars_add = {}
        for var_name, value in vars.items():
            if '_var' in var_name:
                var_name_err = var_name.replace('_var', '_err')
                vars_add[var_name_err] = np.sqrt(vars[var_name])
        vars.update(vars_add)

        vars['DATETIME'] = vars['DATETIME_1'] + (vars['DATETIME_2'] - vars['DATETIME_1'])/2
        vars['T_e'] = vars['T_i'] * vars['T_r']
        vars['T_e_err'] = vars['T_e'] * np.sqrt((vars['T_i_err']/vars['T_i'])**2
                                                + (vars['T_r_err']/vars['T_r'])**2)
        vars['AZ'] = vars['AZ'] % 360.
        # check height and range
        vars['HEIGHT'] = vars['HEIGHT'] / 1000.
        vars['RANGE'] = vars['RANGE'] / 1000.
        inds = np.where(np.isnan(vars['HEIGHT']))
        for i in range(len(inds[0])):
            ind_0 = inds[0][i]
            ind_1 = inds[1][i]
            x0 = np.arange(vars['HEIGHT'].shape[0])
            y0 = vars['HEIGHT'][:, ind_1]
            xp = x0[np.where(np.isfinite(y0))[0]]
            yp = y0[np.isfinite(y0)]
            vars['HEIGHT'][ind_0, ind_1] = np.interp(x0[ind_0], xp, yp)

            x0 = np.arange(vars['RANGE'].shape[0])
            y0 = vars['RANGE'][:, ind_1]
            xp = x0[np.isfinite(y0)]
            yp = y0[np.isfinite(y0)]
            vars['RANGE'][ind_0, ind_1] = np.interp(x0[ind_0], xp, yp)

        if np.isscalar(vars['AZ']):
            az = np.empty((vars['DATETIME_1'].shape[0], 1))
            az[:, :] = vars['AZ']
            vars['AZ'] = az
        if np.isscalar(vars['EL']):
            el = np.empty((vars['DATETIME_1'].shape[0], 1))
            el[:, :] = vars['EL']
            vars['EL'] = el
        vars['AZ'] = np.mod(vars['AZ'], 360)
        vars['EL'] = np.mod(vars['EL'], 360)

        self.variables = vars
        self.metadata = metadata

    def load_eiscat_mat(self):

        raise NotImplemented

    def load_madrigal_hdf5(self):
        from scipy.signal import argrelmin, argrelmax

        var_name_dict = {
            # 'MAGIC_CONSTANT': 'Magic_const',
            'r_SCangle': 'HSA',
            # 'r_m0_2': 'm02',
            # 'r_m0_1': 'm01',
            'AZ': 'AZM',
            'EL': 'ELM',
            'P_Tx': 'POWER',
            'T_SYS_1': 'SYSTMP',
            'T_SYS_2': 'SYSTMP',
            'HEIGHT': 'GDALT',
            'RANGE': 'RANGE',
            'n_e': 'NE',
            'T_i': 'TI',
            'T_r': 'TR',
            'nu_i': 'CO',
            'v_i_los': 'VO',
            'comp_mix': 'PM',
            'comp_O_p': 'PO+',
            'n_e_err': 'DNE',
            'T_i_err': 'DTI',
            'T_r_err': 'DTR',
            'nu_i_err': 'DCO',
            'v_i_los_err': 'DVO',
            'comp_mix_err': 'DPM',
            'comp_O_p_err': 'DPO+',
            'STATUS': 'GFIT',
            'RESIDUAL': 'CHISQ'
        }
        with h5py.File(self.file_path, 'r') as fh5:

            # load metadata
            metadata = {}

            exp_params = fh5['Metadata']['Experiment Parameters'][:]
            exp_params = list(zip(*tuple(exp_params)))
            fn_id = exp_params[0].index(b'Cedar file name')
            fn = exp_params[1][fn_id].decode('UTF-8')
            rc = re.compile(r'^MAD[\w]+_[\d]{4}-[\d]{2}-[\d]{2}_([\w.]+)@(\w+)\.hdf5')
            rm = rc.findall(fn)[0]
            pattern_1 = rm[0]
            pattern_2 = rm[1]
            metadata['modulation'] = pattern_1.split('_')[-1]
            metadata['pulse_code'] = pattern_1.replace('_' + metadata['modulation'], '')

            antenna = pattern_2
            if 'uhf' in antenna:
                sitename = 'UHF'
            elif 'vhf' in antenna:
                sitename = 'VHF'
            elif 'sod' in antenna:
                sitename = 'SOD'
            elif 'kir' in antenna:
                sitename = 'KIR'
            elif '32' in antenna or '42' in antenna:
                sitename = 'ESR'
            else:
                print(antenna)
                raise AttributeError
            metadata['site_name'] = sitename

            lat_id = exp_params[0].index(b'instrument latitude')
            lat = float(exp_params[1][lat_id].decode('UTF-8'))
            lon_id = exp_params[0].index(b'instrument longitude')
            lon = float(exp_params[1][lon_id].decode('UTF-8'))
            alt_id = exp_params[0].index(b'instrument altitude')
            alt = float(exp_params[1][alt_id].decode('UTF-8')) * 1000

            metadata['r_RECloc'] = [lat, lon, alt]
            if sitename in ['ESR', 'UHF', 'VHF', 'TRO']:
                metadata['r_XMITloc'] = [lat, lon, alt]
            else:
                metadata['r_XMITloc'] = [69.583, 19.21, 30]
            metadata['antenna'] = antenna
            metadata['GUISDAP_version'] = ''
            metadata['rawdata_path'] = ''
            metadata['scan_mode'] = ''
            metadata['affiliation'] = ''
            metadata['rawdata_path'] = ''
            metadata['scan_mode'] = ''
            metadata['affiliation'] = ''

            # load data
            data = fh5['Data']['Table Layout'][:]
            data = list(zip(*tuple(data)))
            data_parameters = list(zip(*tuple(fh5['Metadata']['Data Parameters'][:])))
            var_names_h5 = [vn.decode('UTF-8') for vn in data_parameters[0]]
            nvar_h5 = len(var_names_h5)
            ran_id = var_names_h5.index('RANGE')
            ran = data[ran_id]
            inds_ran_min = argrelmin(np.array(ran))[0]
            inds_ran_min = np.append(inds_ran_min, 0)
            inds_ran_min.sort()

            inds_ran_max = argrelmax(np.array(ran))[0]
            inds_ran_max = np.append(inds_ran_max, len(ran)-1)
            inds_ran_max.sort()

            ngates_max = np.max(np.diff(inds_ran_max))
            data_array = np.empty((nvar_h5, inds_ran_min.shape[0], ngates_max))
            data_array[::] = np.nan
            for ip in range(nvar_h5):
                var_tmp = np.array(data[ip])
                for i in range(inds_ran_min.shape[0]):
                    ind1 = inds_ran_min[i]
                    ind2 = inds_ran_max[i]
                    data_array[ip, i, 0: ind2-ind1+1] = var_tmp[ind1: ind2+1]
            num_row = inds_ran_min.shape[0]
            vars_h5 = {}
            for ip in range(nvar_h5):

                var = data_array[ip]
                test_unique = np.unique(var[0][~np.isnan(var[0])])
                if test_unique.shape[0] == 1 and not re.match(r'^D\w+', var_names_h5[ip]) and var_names_h5[ip] not in ['GFIT']:
                    # print(var_names_h5[ip])
                    var = np.array([list(set(var[i][~np.isnan(var[i])])) for i in range(num_row)]).reshape((num_row, 1))
                vars_h5[var_names_h5[ip]] = var

            vars = {}
            for var_name, var_name_h5 in var_name_dict.items():
                try:
                    vars[var_name] = vars_h5[var_name_h5]
                except KeyError:
                    vars[var_name] = None

            num_row = inds_ran_min.shape[0]
            # unix time to datetime
            utime1 = vars_h5['UT1_UNIX'].flatten()
            dt1 = dttool.convert_unix_time_to_datetime_cftime(utime1)
            var = dt1.reshape(num_row, 1)
            var_name = 'DATETIME_1'
            vars[var_name] = var

            utime2 = vars_h5['UT2_UNIX'].flatten()
            dt2 = dttool.convert_unix_time_to_datetime_cftime(utime2)
            var = dt2.reshape(num_row, 1)
            var_name = 'DATETIME_2'
            vars[var_name] = var

            vars['DATETIME'] = vars['DATETIME_1'] + (vars['DATETIME_2'] - vars['DATETIME_1']) / 2
            vars['T_e'] = vars['T_i'] * vars['T_r']
            vars['T_e_err'] = vars['T_e'] * np.sqrt((vars['T_i_err'] / vars['T_i']) ** 2
                                                    + (vars['T_r_err'] / vars['T_r']) ** 2)
            vars['AZ'] = vars['AZ'] % 360.
            inds = np.where(np.isnan(vars['HEIGHT']))
            for i in range(len(inds[0])):
                ind_0 = inds[0][i]
                ind_1 = inds[1][i]
                x0 = np.arange(vars['HEIGHT'].shape[0])
                y0 = vars['HEIGHT'][:, ind_1]
                xp = x0[np.where(np.isfinite(y0))[0]]
                yp = y0[np.isfinite(y0)]
                vars['HEIGHT'][ind_0, ind_1] = np.interp(x0[ind_0], xp, yp)

                x0 = np.arange(vars['RANGE'].shape[0])
                y0 = vars['RANGE'][:, ind_1]
                xp = x0[np.isfinite(y0)]
                yp = y0[np.isfinite(y0)]
                vars['RANGE'][ind_0, ind_1] = np.interp(x0[ind_0], xp, yp)

            self.variables = vars
            self.metadata = metadata
        # raise NotImplemented


if __name__ == "__main__":
    dir_root = prf.datahub_data_root_dir

    # fp = pathlib.Path('examples') / "EISCAT_2005-09-01_steffe_64@32m.hdf5"
    fp = pathlib.Path('examples') / "MAD6400_2021-03-10_beata_ant@uhfa.hdf5"
    Loader(file_path=fp, file_type = 'madrigal-hdf5')



"""
Ovation Prime model modified from Ovation Pyme by lkilcommons: https://github.com/lkilcommons/OvationPyme
Note: this is a new version. Acknowledgement to the authors will be added after the new.
"""
import os
import datetime
from collections import OrderedDict

import numpy as np
from scipy import interpolate

# from ovationprime import ovation_utilities

# from ovationprime.ovation_utilities import robinson_auroral_conductance
# from ovationprime.ovation_utilities import brekke_moen_solar_conductance

# import geospacepy
# from geospacepy import special_datetime, sun, satplottools
import aacgmv2  # available on pip
# import apexpy
from logbook import Logger

log = Logger('OvationPyme.ovation_prime')

# Determine where this module's source file is located
# to determine where to look for the tables
src_file_dir = os.path.dirname(os.path.realpath(__file__))
ovation_datadir = os.path.join(src_file_dir, 'data')


def _check_for_old_jtype(estimator, type_of_flux):
    """Check the type of flux (2nd constructor argument) of
    a FluxEstimator or SeasonalFluxEstimator class and
    raise an extensive error to inform user that they need
    to modify their calling function, and why"""

    name = estimator.__class__.__name__
    explaination = ('Constructor interface to {} has changed'.format(name)
                    + ' now the only valid second argument values'
                    + ' (for type of flux) are "energy" or "number".\n'
                    + ' Formerly, the second argument could take values'
                    + ' which confused types of flux with types of aurora'
                    + ' (e.g. you could specify ion or electron, which'
                    + ' is a property of the auroral type (choose "ion"'
                    + ' auroral type to get ion fluxes).\n'
                    + ' If you wish to calculate average energy, you'
                    + ' will need to switch from a FluxEstimator class'
                    + ' to an AverageEnergyEstimator class')
    if type_of_flux not in ['energy', 'number']:
        raise RuntimeError('{} is not a valid fluxtype.\n{}'.format(type_of_flux,
                                                                    explaination))


class LatLocaltimeInterpolator(object):
    def __init__(self, mlat_grid, mlt_grid, var):
        self.mlat_orig = mlat_grid
        self.mlt_orig = mlt_grid
        self.zvar = var
        n_north, n_south = np.count_nonzero(self.mlat_orig > 0.), np.count_nonzero(self.mlat_orig < 0.)

        if n_south == 0.:
            self.hemisphere = 'N'
        elif n_north == 0.:
            self.hemisphere = 'S'
        else:
            raise ValueError(
                'Latitude grid contains northern (N={0}) and southern (N={1}) values.'.format(n_north, n_south) + \
                ' Can only interpolate one hemisphere at a time.')

    def interpolate(self, new_mlat_grid, new_mlt_grid, method='nearest'):
        """
        Rectangularize and Interpolate (using Linear 2D interpolation)
        """
        X0, Y0 = satplottools.latlt2cart(self.mlat_orig.flatten(), self.mlt_orig.flatten(), self.hemisphere)
        X, Y = satplottools.latlt2cart(new_mlat_grid.flatten(), new_mlt_grid.flatten(), self.hemisphere)
        interpd_zvar = interpolate.griddata((X0, Y0), self.zvar.flatten(), (X, Y), method=method, fill_value=0.)
        return interpd_zvar.reshape(new_mlat_grid.shape)


class BinCorrector(object):
    """
    We've found that often there are strange outlier bins that show up in
    OvationPyme results. This attempts to identify them by computing a numerical
    derivative around each ring of constant latitude.
    """

    def __init__(self, mlat_grid, mlt_grid):
        self.mlat_grid = mlat_grid
        self.mlats = self.mlat_grid[:, 0].flatten()
        self.mlt_grid = mlt_grid
        self.mlts = self.mlt_grid[0, :].flatten()
        self.dy_thresh = None

    def fix(self, y_grid, min_mlat=49, max_mlat=75, label=''):
        """
        Compute derivatives and attempt to identify bad bins
        Assumes mlat varies along the first dimension of the gridded location
        arrays
        """
        debug = False
        plot = False
        bad_bins = np.zeros_like(y_grid, dtype=bool)
        y_grid_corr = y_grid.copy()
        if self.dy_thresh is None:
            self.dy_thresh = 3. * np.nanstd(np.diff(y_grid.flatten()))
        wraparound = lambda x, nwrap: np.concatenate([x[-1 * (nwrap + 1):-1], x, x[:nwrap]])

        for i_mlat, mlat in enumerate(self.mlats):
            if not (np.abs(mlat) >= min_mlat and np.abs(mlat) <= max_mlat):
                if debug:
                    log.debug('MLAT ring at {0} mlat is not between'.format(mlat)
                              + ' {0} and {1}'.format(min_mlat, max_mlat)
                              + ' skipping')
                continue
            mlts_nowrap = self.mlt_grid[i_mlat, :].copy()
            mlts_nowrap[mlts_nowrap < 0] += 24
            mlts_nowrap[-1] = 23.9
            y = y_grid[i_mlat, :]
            # Wrap around first and last nwarp indicies in MLT
            # this prevents out of bounds errors in the spline/derviative
            nwrap = 4  # Pchip is cubic so order+1
            mlts = wraparound(mlts_nowrap, nwrap)
            mlts[:nwrap] -= 24.  # to keep mlt in increasing order
            mlts[-1 * nwrap:] += 24.
            y = wraparound(y, nwrap)
            # y_i = interpolate.PchipInterpolator(mlts, y)
            dy = np.diff(np.concatenate([y[:1], y]))  # compute 1st derivative of spline
            i_dy = interpolate.interp1d(mlts, dy, kind='nearest')
            mlt_mask = np.ones_like(mlts, dtype=bool)
            for i_mlt, mlt in enumerate(mlts_nowrap.flatten()):
                if np.abs(i_dy(mlt)) > self.dy_thresh:
                    bad_bins[i_mlat, i_mlt] = True
                    mlt_mask[i_mlt + nwrap] = False

            y_corr_i = interpolate.PchipInterpolator(mlts[mlt_mask], y[mlt_mask])
            y_corr = y_corr_i(mlts)
            y_grid_corr[i_mlat, :] = y_corr_i(mlts_nowrap)
            if plot:
                self.plot_single_spline(mlat, mlts, y, dy, mlt_mask, y_corr, label=label)

        return y_grid_corr

    def plot_single_spline(self, mlat, mlts, y, dy, mlt_mask, y_corr, label=''):
        import matplotlib.pyplot as plt
        f = plt.figure(figsize=(8, 6))
        ax = f.add_subplot(111)
        ax.plot(mlts, y, 'bo', label='data')
        ax.plot(mlts, dy, 'r-', label='Deriv')
        bad_bins = np.logical_not(mlt_mask)
        ax.plot(mlts, y_corr, 'g.', label='After Correction')
        ax.plot(mlts[bad_bins], y[bad_bins], 'rx',
                label='Bad@dy>{0:.1f}'.format(self.dy_thresh))
        ax.set_title('Spline fit (mlat={0:.1f})'.format(mlat))
        ax.set_xlabel('MLT')
        ax.legend()
        if not os.path.exists('/tmp/ovationpyme'):
            os.makedirs('/tmp/ovationpyme')
        f.savefig('/tmp/ovationpyme/ovationpyme_spline_{0}_{1}.png'.format(label, np.floor(mlat * 10)))
        plt.close(f)

    def __call__(self, y):
        """

        """
        return self.fix(y)


class ConductanceEstimator(object):
    """
    Implements the 'Robinson Formula'
    for estimating Pedersen and Hall height integrated
    conductivity (conducance) from
    average electron energy and
    total electron energy flux
    (assumes a Maxwellian electron energy distribution)
    """

    def __init__(self, fluxtypes=['diff']):

        # Use diffuse aurora only
        self.numflux_estimator = {}
        self.eavg_estimator = {}
        for fluxtype in fluxtypes:
            self.numflux_estimator[fluxtype] = FluxEstimator(fluxtype, 'number')
            self.eavg_estimator[fluxtype] = AverageEnergyEstimator(fluxtype)

    def get_conductance(self, dt, hemi='N', solar=True, auroral=True, background_p=None, background_h=None,
                        conductance_fluxtypes=['diff'], interp_bad_bins=True,
                        return_dF=False, return_f107=False,
                        dnflux_bad_thresh=1.0e8, deavg_bad_thresh=.3):
        """
        Compute total conductance using Robinson formula and emperical solar conductance model
        """
        log.notice(
            "Getting conductance with solar {0}, aurora {1}, fluxtypes {2}, background_ped: {3}, background_hall {4}".format(
                solar,
                auroral, conductance_fluxtypes, background_p, background_h))

        all_sigp_auroral, all_sigh_auroral = [], []
        # Create a bin interpolation corrector
        for fluxtype in conductance_fluxtypes:
            mlat_grid, mlt_grid, numflux_grid, dF = self.numflux_estimator[fluxtype].get_flux_for_time(dt, hemi=hemi,
                                                                                                       return_dF=True)
            # mlat_grid, mlt_grid, energyflux_grid = self.energyflux_estimator.get_flux_for_time(dt, hemi=hemi)
            mlat_grid, mlt_grid, eavg_grid = self.eavg_estimator[fluxtype].get_eavg_for_time(dt, hemi=hemi)

            if interp_bad_bins:
                # Clean up any extremely large bins
                fixer = BinCorrector(mlat_grid, mlt_grid)

                # Fix numflux
                fixer.dy_thresh = dnflux_bad_thresh
                numflux_grid = fixer.fix(numflux_grid, label='nflux_{0}'.format(fluxtype))

                # Fix avg energy
                fixer.dy_thresh = deavg_bad_thresh
                eavg_grid = fixer.fix(eavg_grid, label='eavg_{0}'.format(fluxtype))

                # zero out lowest latitude numflux row because it makes no sense
                # has some kind of artefact at post midnight
                bad = np.abs(mlat_grid) < 52.0
                numflux_grid[bad] = 0.

            # raise RuntimeError('Debug stop!')

            aur_conds = robinson_auroral_conductance(numflux_grid, eavg_grid)
            this_sigp_auroral, this_sigh_auroral = aur_conds
            all_sigp_auroral.append(this_sigp_auroral)
            all_sigh_auroral.append(this_sigh_auroral)

        sigp_solar, sigh_solar, f107 = self.solar_conductance(dt, mlat_grid, mlt_grid, return_f107=True)
        total_sigp_sqrd = np.zeros_like(sigp_solar)
        total_sigh_sqrd = np.zeros_like(sigh_solar)

        if solar:
            total_sigp_sqrd += sigp_solar ** 2
            total_sigh_sqrd += sigh_solar ** 2

        if auroral:
            # Sum up all contributions (sqrt of summed squares)
            for sigp_auroral, sigh_auroral in zip(all_sigp_auroral, all_sigh_auroral):
                total_sigp_sqrd += sigp_auroral ** 2
                total_sigh_sqrd += sigh_auroral ** 2
            # sigp_auroral *= 1.5
            # sigh_auroral *= 1.5

        # Now take square root to get hall and pedersen conductance
        if solar or auroral:
            sigp = np.sqrt(total_sigp_sqrd)
            sigh = np.sqrt(total_sigh_sqrd)
        else:
            # No conductance except flat background
            sigp = total_sigp_sqrd
            sigh = total_sigh_sqrd

        if background_h is not None and background_p is not None:
            # Cousins et. al. 2015, nightside artificial background of 4 Siemens
            # Ellen found this to be the background nightside conductance level which
            # best optimizes the SuperDARN ElePot AMIE ability to predict AMPERE deltaB data, and
            # the AMPERE MagPot AMIE ability to predict SuperDARN LOS V
            sigp[sigp < background_p] = background_p
            sigh[sigh < background_h] = background_h

        if return_dF and return_f107:
            return mlat_grid, mlt_grid, sigp, sigh, dF, f107
        elif return_dF:
            return mlat_grid, mlt_grid, sigp, sigh, dF
        elif return_f107:
            return mlat_grid, mlt_grid, sigp, sigh, f107
        else:
            return mlat_grid, mlt_grid, sigp, sigh

    def solar_conductance(self, dt, mlats, mlts, return_f107=False):
        """
        Estimate the solar conductance using methods from:
            Cousins, E. D. P., T. Matsuo, and A. D. Richmond (2015), Mapping
            high-latitude ionospheric electrodynamics with SuperDARN and AMPERE

            --which cites--

            Asgeir Brekke, Joran Moen, Observations of high latitude ionospheric conductances

            Maybe is not good for SZA for southern hemisphere? Don't know
            Going to use absolute value of latitude because that's what's done
            in Cousins IDL code.
        """
        # Find the closest hourly f107 value
        # to the current time to specifiy the conductance
        f107 = ovation_utilities.get_daily_f107(dt)
        if hasattr(self, '_f107'):
            log.warning(('Warning: Overriding real F107 {0}'.format(f107)
                         + 'with secret instance property _f107 {0}'.format(self._f107)
                         + 'this is for debugging and will not'
                         + 'produce accurate results for a particular date.'))
            f107 = self._f107

        # print "F10.7 = %f" % (f107)

        # Convert from magnetic to geocentric using the AACGMv2 python library
        flatmlats, flatmlts = mlats.flatten(), mlts.flatten()
        flatmlons = aacgmv2.convert_mlt(flatmlts, dt, m2a=True)
        try:
            glats, glons = aacgmv2.convert(flatmlats, flatmlons, 110. * np.ones_like(flatmlats),
                                           date=dt, a2g=True, geocentric=False)
        except AttributeError:
            # convert method was deprecated
            glats, glons, r = aacgmv2.convert_latlon_arr(flatmlats,
                                                         flatmlons,
                                                         110.,
                                                         dt,
                                                         method_code='A2G')

        sigp, sigh = brekke_moen_solar_conductance(dt, glats, glons, f107)

        sigp_unflat = sigp.reshape(mlats.shape)
        sigh_unflat = sigh.reshape(mlats.shape)

        if return_f107:
            return sigp_unflat, sigh_unflat, f107
        else:
            return sigp_unflat, sigh_unflat


class AverageEnergyEstimator(object):
    """A class which estimates average energy by estimating both
    energy and number flux
    """

    def __init__(self, atype, numflux_threshold=5.0e7):
        self.numflux_threshold = numflux_threshold
        self.numflux_estimator = FluxEstimator(atype, 'number')
        self.energyflux_estimator = FluxEstimator(atype, 'energy')

    def get_eavg_for_time(self, dt, hemi='N', return_dF=False, combine_hemispheres=True):

        kwargs = {
            'hemi': hemi,
            'combine_hemispheres': combine_hemispheres,
            'return_dF': True
        }

        grid_mlats, grid_mlts, gridnumflux, dF = self.numflux_estimator.get_flux_for_time(dt, **kwargs)
        grid_mlats, grid_mlts, gridenergyflux, dF = self.energyflux_estimator.get_flux_for_time(dt, **kwargs)

        grideavg = (gridenergyflux / 1.6e-12) / gridnumflux  # energy flux Joules->eV
        grideavg = grideavg / 1000.  # eV to keV

        # Limit to reasonable number fluxes
        n_pts = len(grideavg.flatten())
        n_low_numflux = np.count_nonzero(gridnumflux < self.numflux_threshold)
        grideavg[gridnumflux < self.numflux_threshold] = 0.
        log.debug(('Zeroed {:d}/{:d} average energies'.format(n_low_numflux, n_pts)
                   + 'with numflux below {:e}'.format(self.numflux_threshold)))

        # Limit to DMSP SSJ channels range
        n_over = np.count_nonzero(grideavg > 30)
        n_under = np.count_nonzero(grideavg < .5)
        log.debug('Zeroed {:d}/{:d} average energies over 30 keV'.format(n_over, n_pts))
        log.debug('Zeroed {:d}/{:d} average energies under .2 keV'.format(n_under, n_pts))
        grideavg[grideavg > 30.] = 30.  # Max of 30keV
        grideavg[grideavg < .2] = 0.  # Min of 1 keV

        if not return_dF:
            return grid_mlats, grid_mlts, grideavg
        else:
            return grid_mlats, grid_mlts, grideavg, dF


class FluxEstimator(object):
    """
    A class which estimates auroral flux
    based on the Ovation Prime regressions,
    at arbitrary locations and times.

    Locations are in magnetic latitude and local
    time, and are interpolated using a B-spline
    representation
    """

    def __init__(self, atype, energy_or_number, seasonal_estimators=None):
        """

        doy - int
            day of year

        atype - str, ['diff','mono','wave','ions']
            type of aurora for which to load regression coeffients

        energy_or_number - str, ['energy','number']

            Type of flux you want to estimate

        seasonal_estimators - dict, optional
            A dictionary of SeasonalFluxEstimators for seasons
            'spring','fall','summer','winter', if you
            don't want to create them
            (for efficiency across multi-day calls)

        """
        self.atype = atype  # Type of aurora

        # Check for legacy values of this argument
        _check_for_old_jtype(self, energy_or_number)

        self.energy_or_number = energy_or_number  # Type of flux

        seasons = ['spring', 'summer', 'fall', 'winter']

        if seasonal_estimators is None:
            # Make a seasonal estimator for each season with nonzero weight
            self.seasonal_flux_estimators = {season: SeasonalFluxEstimator(season, atype, energy_or_number) for season
                                             in seasons}
        else:
            # Ensure the passed seasonal estimators are approriate for this atype and jtype
            for season, estimator in seasonal_estimators.items():
                jtype_atype_ok = jtype_atype_ok and (self.jtype == estimator.jtype and self.atype == estimator.atype)
            if not jtype_atype_ok:
                raise RuntimeError(
                    'Auroral and flux type of SeasonalFluxEstimators do not match {0} and {1}!'.format(self.atype,
                                                                                                       self.jtype))

    def season_weights(self, doy):
        """
        Determines the relative weighting of the
        model coeffecients for the various seasons for a particular
        day of year (doy). Nominally, weights the seasons
        based on the difference between the doy and the peak
        of the season (solstice/equinox)

        Returns:
            a dictionary with a key for each season.
            Each value in the dicionary is a float between 0 and 1
        """
        weight = OrderedDict(winter=0.,
                             spring=0.,
                             summer=0.,
                             fall=0.)

        if doy >= 79. and doy < 171:
            weight['summer'] = 1. - (171. - doy) / 92.
            weight['spring'] = 1. - weight['summer']

        elif doy >= 171. and doy < 263.:
            weight['fall'] = 1. - (263. - doy) / 92.
            weight['summer'] = 1. - weight['fall']

        elif doy >= 263. and doy < 354.:
            weight['winter'] = 1. - (354. - doy) / 91.
            weight['fall'] = 1. - weight['winter']

        elif doy >= 354 or doy < 79:
            # For days of year > 354, subtract 365 to get negative
            # day of year values for computation
            doy0 = doy - 365. if doy >= 354 else doy
            weight['spring'] = 1. - (79. - doy0) / 90.
            weight['winter'] = 1. - weight['spring']

        return weight

    def get_season_fluxes(self, dF, weights):
        """
        Extract the flux for each season and hemisphere and
        store them in a dictionary
        Return positive latitudes, since northern and southern
        latitude/localtime grids are the same
        """
        seasonfluxesN, seasonfluxesS = OrderedDict(), OrderedDict()
        gridmlats, gridmlts = None, None
        for season, estimator in self.seasonal_flux_estimators.items():
            if weights[season] == 0.:
                continue  # Skip calculation for seasons with zero weight

            flux_outs = estimator.get_gridded_flux(dF)
            gridmlatsN, gridmltsN, gridfluxN = flux_outs[:3]
            gridmlatsS, gridmltsS, gridfluxS = flux_outs[3:]
            seasonfluxesN[season] = gridfluxN
            seasonfluxesS[season] = gridfluxS
            gridmlats = gridmlatsN
            gridmlts = gridmltsN
        return gridmlats, gridmlts, seasonfluxesN, seasonfluxesS

    def get_flux_for_time(self, dt,
                          hemi='N', return_dF=False, combine_hemispheres=True):
        """
        The weighting of the seasonal flux for the different hemispheres
        is a bit counterintuitive, but after some investigation of the flux
        patterns produced for Northern and Southern hemispheres using a
        particular SeasonalFluxEstimator (which in turn reads a particular
        season's coefficients file), it seems like the 'summer' coefficients
        file contains the northern hemisphere coefficients for Boreal Summer
        (roughly May-August) and the southern hemisphere coefficients for
        Austral Summer (roughly November-February).

        In earlier versions of this code, the season weighting was wrong,
        because the code operated on the assumption that 'summer'
        meant Boreal summer, and specified a range of dates for the data
        used to construct the coefficients.

        In the IDL version of this model, the flux produced for
        Northern and Southern hemispheres is averaged, with the following
        comment on the IDL keyword argument:

        ;n_or_s=3 for combine north and south.  In effect this is the only
        ;option that works.  The result is appropriate to the northern
        ;hemisphere.  To get a result appropriate to the southern hemisphere,
        ;call with doy = 365 - actual doy

        Combining hemispheres is probably nessecary because
        there are data gaps (particularly in the northern hemisphere dawn)
        so this is the default behavior here as well. This can be overriden
        by passing combine_hemispheres=False
        """
        doy = dt.timetuple().tm_yday

        if not combine_hemispheres:
            log.warning(('Warning: IDL version of OP2010 always combines hemispheres.'
                         + 'know what you are doing before switching this behavior'))

        if hemi == 'N':
            weights = self.season_weights(doy)
        elif hemi == 'S':
            weights = self.season_weights(365. - doy)
        else:
            raise ValueError('Invalid hemisphere {0} (use N or S)'.format(hemi))

        dF = ovation_utilities.calc_dF(dt)
        if hasattr(self, '_dF'):
            log.warning(('Warning: Overriding real Newell Coupling {0}'.format(dF)
                         + 'with secret instance property _dF {0}'.format(self._dF)
                         + 'this is for debugging and will not'
                         + 'produce accurate results for a particular date'))
            dF = self._dF

        season_fluxes_outs = self.get_season_fluxes(dF, weights)
        grid_mlats, grid_mlts, seasonfluxesN, seasonfluxesS = season_fluxes_outs

        gridflux = np.zeros_like(grid_mlats)
        for season, W in weights.items():
            if W == 0.:
                continue

            gridfluxN = seasonfluxesN[season]
            gridfluxS = seasonfluxesS[season]

            if combine_hemispheres:
                gridflux += W * (gridfluxN + gridfluxS) / 2
            elif hemi == 'N':
                gridflux += W * gridfluxN
            elif hemi == 'S':
                gridflux += W * gridfluxS

        if hemi == 'S':
            grid_mlats = -1. * grid_mlats  # by default returns positive latitudes

        if not return_dF:
            return grid_mlats, grid_mlts, gridflux
        else:
            return grid_mlats, grid_mlts, gridflux, dF


class SeasonalFluxEstimator(object):
    """
    A class to hold and caculate predictions from the regression coeffecients
    which are tabulated in the data/premodel/{season}_{atype}_*.txt
    files.

    Given a particular season, type of aurora ( one of ['diff','mono','wave'])
    and type of flux, returns
    """

    _valid_atypes = ['diff', 'mono', 'wave', 'ions']

    def __init__(self, season, atype, energy_or_number):
        """
        season - str,['winter','spring','summer','fall']
            season for which to load regression coeffients

        atype - str, ['diff','mono','wave','ions']
            type of aurora for which to load regression coeffients, ions
            are not implemented

        energy_or_number - str, ['energy','number']
            type of flux you want to estimate
        """

        nmlt = 96  # number of mag local times in arrays (resolution of 15 minutes)
        nmlat = 160  # number of mag latitudes in arrays (resolution of 1/4 of a degree (.25))
        ndF = 12  # number of coupling strength bins

        self.n_mlt_bins, self.n_mlat_bins, self.n_dF_bins = nmlt, nmlat, ndF

        self.atype = atype
        if atype not in self._valid_atypes:
            raise ValueError(('Not a valid aurora type {}.'.format(atype)
                              + 'valid values {}'.format(self._valid_atypes)))

        # Check for legacy values of this argument
        _check_for_old_jtype(self, energy_or_number)

        self.energy_or_number = energy_or_number

        # The mlat bins are orgainized like -50:-dlat:-90, 50:dlat:90
        self.mlats = np.concatenate([np.linspace(-90., -50., self.n_mlat_bins // 2)[::-1],
                                     np.linspace(50., 90., self.n_mlat_bins // 2)])

        self.mlts = np.linspace(0., 24., self.n_mlt_bins)

        # Determine file names
        file_suffix = '_n' if energy_or_number == 'number' else ''
        self.afile = os.path.join(ovation_datadir, 'premodel/{0}_{1}{2}.txt'.format(season, atype, file_suffix))
        self.pfile = os.path.join(ovation_datadir, 'premodel/{0}_prob_b_{1}.txt'.format(season, atype))
        # Defualt values of header (don't know why need yet)
        # b1 = 0.
        # b2 = 0.
        # yend = 1900
        # dend = 1
        # y0 = 1900
        # d0 = 1
        # files_done = 0
        # sf0 = 0

        with open(self.afile, 'r') as f:
            aheader = f.readline()  # y0,d0,yend,dend,files_done,sf0
            # print "Read Auroral Flux Coefficient File %s,\n Header: %s" % (self.afile,aheader)
            # Don't know if it will read from where f pointer is after reading header line
            adata = np.genfromtxt(f, max_rows=nmlat * nmlt)
            # print "First line was %s" % (str(adata[0,:]))

        # These are the coefficients for each bin which are used
        # in the predicted flux calulation for electron auroral types
        # and for ions
        self.b1a, self.b2a = np.zeros((nmlt, nmlat)), np.zeros((nmlt, nmlat))
        self.b1a.fill(np.nan)
        self.b2a.fill(np.nan)
        mlt_bin_inds, mlat_bin_inds = adata[:, 0].astype(int), adata[:, 1].astype(int)
        self.b1a[mlt_bin_inds, mlat_bin_inds] = adata[:, 2]
        self.b2a[mlt_bin_inds, mlat_bin_inds] = adata[:, 3]

        self.b1p = np.full((nmlt, nmlat), np.nan)
        self.b2p = np.full((nmlt, nmlat), np.nan)
        self.prob = np.full((nmlt, nmlat, ndF), np.nan)

        # pdata has 2 columns, b1, b2 for first 15361 rows
        # pdata has nmlat*nmlt rows (one for each positional bin)

        # Electron auroral types also include a probability in their
        # predicted flux calculations (related to the probability of
        # observing one type of aurora versus another)
        if atype in ['diff', 'mono', 'wave']:
            with open(self.pfile, 'r') as f:
                pheader = f.readline()  # y0,d0,yend,dend,files_done,sf0
                # Don't know if it will read from where f pointer is after reading header line
                pdata_b = np.genfromtxt(f, max_rows=nmlt * nmlat)  # 2 columns, b1 and b2
                # print "Shape of b1p,b2p should be nmlt*nmlat=%d, is %s" % (nmlt*nmlat,len(pdata_b[:,0]))
                pdata_p = np.genfromtxt(f, max_rows=nmlt * nmlat * ndF)  # 1 column, pval

            # in the file the probability is stored with coupling strength bin
            # varying fastest (this is Fortran indexing order)
            pdata_p_column_dFbin = pdata_p.reshape((-1, ndF), order='F')

            # mlt is first dimension
            self.b1p[mlt_bin_inds, mlat_bin_inds] = pdata_b[:, 0]
            self.b2p[mlt_bin_inds, mlat_bin_inds] = pdata_b[:, 1]
            for idF in range(ndF):
                self.prob[mlt_bin_inds, mlat_bin_inds, idF] = pdata_p_column_dFbin[:, idF]

        # IDL original read
        # readf,20,i,j,b1,b2,rF
        # ;;   b1a_all(atype, iseason,i,j) = b1
        # ;;   b2a_all(atype, iseason,i,j) = b2
        # adata has 5 columns, mlt bin number, mlat bin number, b1, b2, rF
        # adata has nmlat*nmlt rows (one for each positional bin)

    def which_dF_bin(self, dF):
        """
        Given a coupling strength value, finds the bin it falls into
        """
        dFave = 4421.  # Magic numbers!
        dFstep = dFave / 8.
        i_dFbin = np.round(dF / dFstep)
        # Range check 0 <= i_dFbin <= n_dF_bins-1
        if i_dFbin < 0 or i_dFbin > self.n_dF_bins - 1:
            i_dFbin = 0 if i_dFbin < 0 else self.n_dF_bins - 1
        return int(i_dFbin)

    def prob_estimate(self, dF, i_mlt_bin, i_mlat_bin):
        """
        Estimate probability of <something> by using tabulated
        linear regression coefficients ( from prob_b files )
        WRT coupling strength dF (which are different for each position bin)

        If p doesn't come out sensible by the initial regression,
        (i.e both regression coefficients are zero)
        then tries loading from the probability array. If the value
        in the array is zero, then estimates a value using adjacent
        coupling strength bins in the probability array
        """

        # Look up the regression coefficients
        b1, b2 = self.b1p[i_mlt_bin, i_mlat_bin], self.b2p[i_mlt_bin, i_mlat_bin]

        p = b1 + b2 * dF  # What is this the probability of?

        # range check 0<=p<=1
        if p < 0. or p > 1.:
            p = 1. if p > 1. else 0.

        if b1 == 0. and b2 == 0.:
            i_dFbin = self.which_dF_bin(dF)
            # Get the tabulated probability
            p = self.prob[i_mlt_bin, i_mlat_bin, i_dFbin]

            if p == 0.:
                # If no tabulated probability we must estimate by interpolating
                # between adjacent coupling strength bins
                i_dFbin_1 = i_dFbin - 1 if i_dFbin > 0 else i_dFbin + 2  # one dF bin before by preference, two after in extremis
                i_dFbin_2 = i_dFbin + 1 if i_dFbin < self.n_dF_bins - 1 else i_dFbin - 2  # one dF bin after by preference, two before in extremis
                p = (self.prob[i_mlt_bin, i_mlat_bin, i_dFbin_1] + self.prob[i_mlt_bin, i_mlat_bin, i_dFbin_2]) / 2.

        return p

    def estimate_auroral_flux(self, dF, i_mlt_bin, i_mlat_bin):
        """
        Does what it says on the tin,
        estimates the flux using the regression coeffecients in the 'a' files
        """
        b1, b2 = self.b1a[i_mlt_bin, i_mlat_bin], self.b2a[i_mlt_bin, i_mlat_bin]
        # There are no spectral types for ions, so there is no need
        # to weight the predicted flux by a probability
        if self.atype == 'ions':
            p = 1.
        else:
            p = self.prob_estimate(dF, i_mlt_bin, i_mlat_bin)
        # print(p, b1, b2, dF)
        flux = (b1 + b2 * dF) * p
        return self.correct_flux(flux)

    def correct_flux(self, flux):
        """
        A series of magical (unexplained, unknown) corrections to flux given a particular
        type of flux
        """
        fluxtype = self.energy_or_number

        if flux < 0.:
            flux = 0.

        if self.atype is not 'ions':
            # Electron Energy Flux
            if fluxtype == 'energy':
                if flux > 10.:
                    flux = 0.5
                elif flux > 5.:
                    flux = 5.

            # Electron Number Flux
            elif fluxtype == 'number':
                if flux > 2.0e9:
                    flux = 1.0e9
                elif flux > 2.0e10:
                    flux = 0.
        else:
            # Ion Energy Flux
            if fluxtype == 'energy':
                if flux > 2.:
                    flux = 2.
                elif flux > 4.:
                    flux = 0.25

            # Ion Number Flux
            if fluxtype == 'number':
                if flux > 1.0e8:
                    flux = 1.0e8
                elif flux > 5.0e8:
                    flux = 0.
        return flux

    def get_gridded_flux(self, dF, combined_N_and_S=False, interp_N=True):
        """
        Return the flux interpolated onto arbitary locations
        in mlats and mlts

        combined_N_and_S, bool, optional
            Average the fluxes for northern and southern hemisphere
            and use them for both hemispheres (this is what standard
            ovation prime does always I think, so I've made it default)
            The original code says that this result is appropriate for
            the northern hemisphere, and to use 365 - actual doy to
            get a basic result appropriate for the southern hemisphere

        interp_N, bool, optional
            Interpolate flux linearly for each latitude ring in the wedge
            of low coverage in northern hemisphere dawn/midnight region
        """

        fluxgridN = np.zeros((self.n_mlat_bins // 2, self.n_mlt_bins))
        fluxgridN.fill(np.nan)
        # Make grid coordinates
        mlatgridN, mltgridN = np.meshgrid(self.mlats[self.n_mlat_bins // 2:], self.mlts, indexing='ij')

        fluxgridS = np.zeros((self.n_mlat_bins // 2, self.n_mlt_bins))
        fluxgridS.fill(np.nan)
        # Make grid coordinates
        mlatgridS, mltgridS = np.meshgrid(self.mlats[:self.n_mlat_bins // 2], self.mlts, indexing='ij')
        # print(self.mlats[:self.n_mlat_bins//2])

        for i_mlt in range(self.n_mlt_bins):
            for j_mlat in range(self.n_mlat_bins // 2):
                # The mlat bins are orgainized like -50:-dlat:-90,50:dlat:90
                fluxgridN[j_mlat, i_mlt] = self.estimate_auroral_flux(dF, i_mlt, self.n_mlat_bins // 2 + j_mlat)
                fluxgridS[j_mlat, i_mlt] = self.estimate_auroral_flux(dF, i_mlt, j_mlat)

        if interp_N:
            fluxgridN, inwedge = self.interp_wedge(mlatgridN, mltgridN, fluxgridN)
            self.inwedge = inwedge

        if not combined_N_and_S:
            return mlatgridN, mltgridN, fluxgridN, mlatgridS, mltgridS, fluxgridS
        else:
            return mlatgridN, mltgridN, (fluxgridN + fluxgridS) / 2.

    def interp_wedge(self, mlatgridN, mltgridN, fluxgridN):
        """
        Interpolates across the wedge shaped data gap
        around 50 magnetic latitude and 23-4 MLT.
        Interpolation is performed individually
        across each magnetic latitude ring,
        only missing flux values are filled with the
        using the interpolant
        """
        # Constants copied verbatim from IDL code
        x_mlt_min = -1.0  # minimum MLT for interpolation [hours] --change if desired
        x_mlt_max = 4.0  # maximum MLT for interpolation [hours] --change if desired
        x_mlat_min = 49.0  # minimum MLAT for interpolation [degrees]
        # x_mlat_max=67.0
        x_mlat_max = 75.0  # maximum MLAT for interpolation [degrees] --change if desired (LMK increased this from 67->75)

        valid_interp_mlat_bins = np.logical_and(mlatgridN[:, 0] >= x_mlat_min, mlatgridN[:, 0] <= x_mlat_max).flatten()
        inwedge = np.zeros(fluxgridN.shape, dtype=bool)  # Store where we did interpolation

        for i_mlat_bin in np.flatnonzero(valid_interp_mlat_bins).tolist():
            # Technically any row in the MLT grid would do, but for consistancy use the i_mlat_bin-th one
            this_mlat = mlatgridN[i_mlat_bin, 0].copy()
            this_mlt = mltgridN[i_mlat_bin, :].copy()
            this_flux = fluxgridN[i_mlat_bin, :].copy()

            # Change from 0-24 MLT to -12 to 12 MLT, so that there is no discontiunity at midnight
            # when we interpolate
            this_mlt[this_mlt > 12.] = this_mlt[this_mlt > 12.] - 24.

            valid_interp_mlt_bins = np.logical_and(this_mlt >= x_mlt_min, this_mlt <= x_mlt_max).flatten()
            mlt_bins_missing_flux = np.logical_not(this_flux > 0.).flatten()

            interp_bins_missing_flux = np.logical_and(valid_interp_mlt_bins, mlt_bins_missing_flux)

            inwedge[i_mlat_bin, :] = interp_bins_missing_flux

            if np.count_nonzero(interp_bins_missing_flux) > 0:
                # Bins right next to missing wedge probably have bad statistics, so
                # don't include them
                interp_bins_missing_flux_inds = np.flatnonzero(interp_bins_missing_flux)
                nedge = 6
                for edge_offset in range(1, nedge + 1):
                    lower_edge_ind = interp_bins_missing_flux_inds[0] - edge_offset
                    upper_edge_ind = np.mod(interp_bins_missing_flux_inds[-1] + edge_offset,
                                            len(interp_bins_missing_flux))
                    interp_bins_missing_flux[lower_edge_ind] = interp_bins_missing_flux[
                        interp_bins_missing_flux_inds[0]]
                    interp_bins_missing_flux[upper_edge_ind] = interp_bins_missing_flux[
                        interp_bins_missing_flux_inds[-1]]

                interp_source_bins = np.flatnonzero(np.logical_not(interp_bins_missing_flux))

                # flux_interp = interpolate.PchipInterpolator(this_mlt[interp_source_bins], this_flux[interp_source_bins])
                flux_interp = interpolate.interp1d(this_mlt[interp_source_bins], this_flux[interp_source_bins],
                                                   kind='linear')
                fluxgridN[i_mlat_bin, interp_bins_missing_flux] = flux_interp(this_mlt[interp_bins_missing_flux])

                # print fluxgridN[i_mlat_bin,interp_bins_missing_flux]
                # print "For latitude %.1f, replaced %d flux bins between MLT %.1f and %.1f with interpolated flux..." % (this_mlat,
                #       np.count_nonzero(interp_bins_missing_flux),np.nanmin(this_mlt[interp_bins_missing_flux]),
                #       np.nanmax(this_mlt[interp_bins_missing_flux]))

        # notwedge = np.logical_not(inwedge)
        # crazy = np.logical_and(inwedge,fluxgridN>np.nanpercentile(fluxgridN[notwedge],90.))
        # fluxgridN[crazy]=np.nanpercentile(fluxgridN[notwedge],90.)

        return fluxgridN, inwedge




import numpy as np
from scipy.interpolate import interp1d
import os
import datetime
from scipy.interpolate import griddata
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from cartopy.mpl.ticker import (
    LongitudeLocator, LatitudeLocator,
    LongitudeFormatter, LatitudeFormatter)
import matplotlib.ticker as mticker
import matplotlib.path as mpath
import matplotlib.cm as cm
from loaders.load_uta_gitm_201602_newrun import *
import utilities.datetime_utilities as du

import visualization.time_series as ts
import scipy.signal as sig

from geospacelab.visualization.map_proj.geopanel import PolarView


def get_swarm_data(dt_fr, dt_to, satID="C"):
    dt_range = [dt_fr, dt_to]

    instr_info1 = {'name': 'SWARM-' + satID, 'assign_key': 'SWARM_ACC'}

    database = 'uta'

    paralist = [
        {'database': database, 'instrument': instr_info1, 'paraname': 'N_n_SC'},
        {'database': database, 'instrument': instr_info1, 'paraname': 'N_n_s450'},
        {'database': database, 'instrument': instr_info1, 'paraname': 'N_n_s500'},
    ]

    paras_layout = [[1, 2]]

    tsObj = ts.TimeSeries(dt_range=dt_range, paralist=paralist, paras_layout=paras_layout, timeline='multiple')

    sc_lat = tsObj.dataObjs['uta_swarm_c_acc'].paras['SC_GEO_LAT']
    sc_lon = tsObj.dataObjs['uta_swarm_c_acc'].paras['SC_GEO_LON']
    sc_lst = tsObj.dataObjs['uta_swarm_c_acc'].paras['SC_GEO_ST']
    sc_datetime = tsObj.dataObjs['uta_swarm_c_acc'].paras['SC_DATETIME']
    rho_n_sc = tsObj.dataObjs['uta_swarm_c_acc'].paras['N_n_SC']

    swarm_data = {
        'sc_lat': sc_lat,
        'sc_lon': sc_lon,
        'sc_lst': sc_lst,
        'sc_datetime': sc_datetime,
        'rho_n_sc': rho_n_sc
    }

    return swarm_data


def show_rho_n(dt_fr, dt_to):
    swarm_data = get_swarm_data(dt_fr, dt_to)
    sc_lon = swarm_data['sc_lon'].flatten()
    sc_lat = swarm_data['sc_lat'].flatten()
    sc_dt = swarm_data['sc_datetime'].flatten()
    rho_n_sc = swarm_data['rho_n_sc'].flatten()

    plt.figure(figsize=(8,8))
    panel = PolarView(cs='GEO', pole='N', lon_c=None, lst_c=0, ut=dt_fr, boundary_lat=0., proj_style='Stereographic')
    panel.add_subplot(major=True)
    panel.set_extent(boundary_style='circle')

    data = panel.proj.transform_points(ccrs.PlateCarree(), sc_lon, sc_lat)

    x = data[:, 0]
    y = data[:, 1]


    from scipy.stats import linregress
    coef = linregress(x, y)
    a = coef.slope
    b = coef.intercept

    # x1 = np.linspace(np.nanmin(x), np.nanmax(x), num=500)
    # y1 = np.linspace(np.nanmin(y), np.nanmax(y), num=500)
    # x2d, y2d = np.meshgrid(x, y)
    # z2d = griddata(data[:, 0:2], rho_n_sc.flatten(), (x2d, y2d), method='nearest')
    # z2d_dist = np.abs(a*x2d - y2d + b) / np.sqrt(a**2 + 1)
    # z2d = np.where(z2d_dist>1000, np.nan, z2d)
    #
    # im = panel.major_ax.pcolormesh(x2d, y2d, z2d, vmin=2e-14, vmax=20e-13, cmap='gist_ncar')

    #  panel.major_ax.plot(sc_lon, sc_lat, transform=ccrs.Geodetic(), linewidth=0.5, color='k')
    # plt.colorbar(im)

    # xx = np.tile(x, [3, 1])
    #
    # yy = np.concatenate((y[np.newaxis, :]-150000, y[np.newaxis, :], y[np.newaxis, :]+150000), axis=0)
    # zz = np.concatenate((rho_n_sc.T, rho_n_sc.T, rho_n_sc.T), axis=0)
    #
    # zz = griddata(data[:, 0:2], rho_n_sc.flatten(), (xx, yy), method='nearest')
    # im = panel.major_ax.pcolormesh(xx, yy, zz, vmin=2e-14, vmax=20e-13, cmap='gist_ncar')
    #
    # #  panel.major_ax.plot(sc_lon, sc_lat, transform=ccrs.Geodetic(), linewidth=0.5, color='k')
    # plt.colorbar(im)

    #panel.major_ax.plot(data[:,0], data[:,1], linewidth=3)

    from matplotlib.collections import LineCollection
    from matplotlib.colors import ListedColormap, BoundaryNorm

    data = panel.proj.transform_points(ccrs.PlateCarree(), sc_lon, sc_lat)
    x = data[:, 0]
    y = data[:, 1]
    z = rho_n_sc.flatten()
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    norm = plt.Normalize(2e-14, 18e-13)
    lc = LineCollection(segments, cmap='gist_ncar', norm=norm)
    lc.set_array(z)
    lc.set_linewidth(6)
    line = panel.major_ax.add_collection(lc)
    cbar = plt.gcf().colorbar(line, ax=panel.major_ax, pad=0.1, fraction=0.03)
    cbar.set_label('Neutral mass density\n' + r'(kg/m$^{3}$)', rotation=270, labelpad=25)
    #cbaxes = plt.gcf().add_axes([0.8, 0.1, 0.03, 0.8]) 
    #cb = plt.colorbar(panel.major_ax, cax = cbaxes)  

    sectime, dt0 = du.convert_datetime_to_sectime(sc_dt, datetime.datetime(dt_fr.year, dt_fr.month, dt_fr.day))
    sectime_res = 10 * 60
    time_ticks = np.arange(np.floor(sectime[0]/sectime_res)*sectime_res, np.ceil(sectime[-1]/sectime_res)*sectime_res, sectime_res)
    from scipy.interpolate import interp1d
    f = interp1d(sectime, x, fill_value='extrapolate')
    x_time_ticks = f(time_ticks)
    f = interp1d(sectime, y, fill_value='extrapolate')
    y_time_ticks = f(time_ticks)
    panel.major_ax.plot(x_time_ticks, y_time_ticks, '.', markersize=4, color='k')
    for ind, time_tick in enumerate(time_ticks):
        time = dt0 + datetime.timedelta(seconds=time_tick)
        x_time_tick = x_time_ticks[ind]
        y_time_tick = y_time_ticks[ind]

        if x_time_tick < panel._extent[0] or x_time_tick > panel._extent[1]:
            continue
        if y_time_tick < panel._extent[2] or y_time_tick > panel._extent[3]:
            continue

        panel.major_ax.text( x_time_tick, y_time_tick, time.strftime("%d %H:%M"), fontsize=6)

    panel.add_coastlines()
    panel.add_grids()

    plt.gcf().suptitle('Swarm-C neutral mass density\n 2016-02-03T07:00 - 2016-02-03T07:50')
    plt.savefig('test_pho_n.png', dpi=300)
    # plt.show()

def show_n_e(dt_fr, dt_to):
    import cdflib
    filepath = "~/tmp/SW_OPER_EFIC_LP_1B_20160203T000000_20160203T235959_0501_MDR_EFI_LP.cdf"
    cf = cdflib.CDF(filepath)
    cf_info = cf.cdf_info()
    n_e = cf.varget(variable='Ne')
    T_e = cf.varget(variable='Te')
    sc_lat = cf.varget(variable='Latitude')
    sc_lon = cf.varget(variable='Longitude')
    timestamp = cf.varget(variable='Timestamp')
    
    dtstrs = cdflib.cdfepoch.encode(timestamp)
    dts = np.empty_like(timestamp, dtype=datetime.datetime)
    for ind, dtstr in enumerate(dtstrs):
        dts[ind] = datetime.datetime.strptime(dtstr+'000', '%Y-%m-%dT%H:%M:%S.%f')
    ind_dt = np.where((dts >= dt_fr) & (dts <= dt_to))[0]
    # times = cdflib.cdfepoch.unixtime(timestamp, to_np=True)
    
    
    sc_lon = sc_lon[ind_dt]
    sc_lat = sc_lat[ind_dt]
    sc_dt = dts[ind_dt]
    rho_n_sc = n_e[ind_dt]

    plt.figure(figsize=(8,8))
    panel = PolarView(cs='GEO', pole='N', lon_c=None, lst_c=0, ut=dt_fr, boundary_lat=0., proj_style='Stereographic')
    panel.add_subplot(major=True)
    panel.set_extent(boundary_style='circle')

    data = panel.proj.transform_points(ccrs.PlateCarree(), sc_lon, sc_lat)

    x = data[:, 0]
    y = data[:, 1]


    from scipy.stats import linregress
    coef = linregress(x, y)
    a = coef.slope
    b = coef.intercept

    # x1 = np.linspace(np.nanmin(x), np.nanmax(x), num=500)
    # y1 = np.linspace(np.nanmin(y), np.nanmax(y), num=500)
    # x2d, y2d = np.meshgrid(x, y)
    # z2d = griddata(data[:, 0:2], rho_n_sc.flatten(), (x2d, y2d), method='nearest')
    # z2d_dist = np.abs(a*x2d - y2d + b) / np.sqrt(a**2 + 1)
    # z2d = np.where(z2d_dist>1000, np.nan, z2d)
    #
    # im = panel.major_ax.pcolormesh(x2d, y2d, z2d, vmin=2e-14, vmax=20e-13, cmap='gist_ncar')

    #  panel.major_ax.plot(sc_lon, sc_lat, transform=ccrs.Geodetic(), linewidth=0.5, color='k')
    # plt.colorbar(im)

    # xx = np.tile(x, [3, 1])
    #
    # yy = np.concatenate((y[np.newaxis, :]-150000, y[np.newaxis, :], y[np.newaxis, :]+150000), axis=0)
    # zz = np.concatenate((rho_n_sc.T, rho_n_sc.T, rho_n_sc.T), axis=0)
    #
    # zz = griddata(data[:, 0:2], rho_n_sc.flatten(), (xx, yy), method='nearest')
    # im = panel.major_ax.pcolormesh(xx, yy, zz, vmin=2e-14, vmax=20e-13, cmap='gist_ncar')
    #
    # #  panel.major_ax.plot(sc_lon, sc_lat, transform=ccrs.Geodetic(), linewidth=0.5, color='k')
    # plt.colorbar(im)

    #panel.major_ax.plot(data[:,0], data[:,1], linewidth=3)

    from matplotlib.collections import LineCollection
    from matplotlib.colors import ListedColormap, BoundaryNorm

    data = panel.proj.transform_points(ccrs.PlateCarree(), sc_lon, sc_lat)
    x = data[:, 0]
    y = data[:, 1]
    z = rho_n_sc.flatten()
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    norm = plt.Normalize(1e4, 1e6)
    norm = colors.LogNorm(vmin=1e4, vmax=1e6)
    lc = LineCollection(segments, cmap='gist_ncar', norm=norm)
    lc.set_array(z)
    lc.set_linewidth(6)
    line = panel.major_ax.add_collection(lc)
    cbar = plt.gcf().colorbar(line, ax=panel.major_ax, pad=0.1, fraction=0.03)
    cbar.set_label(r'$n_e$' + '\n' + r'(cm$^{-3}$)', rotation=270, labelpad=25)
    #cbaxes = plt.gcf().add_axes([0.8, 0.1, 0.03, 0.8]) 
    #cb = plt.colorbar(panel.major_ax, cax = cbaxes)  


    sectime, dt0 = du.convert_datetime_to_sectime(sc_dt, datetime.datetime(dt_fr.year, dt_fr.month, dt_fr.day))
    sectime_res = 10 * 60
    time_ticks = np.arange(np.floor(sectime[0]/sectime_res)*sectime_res, np.ceil(sectime[-1]/sectime_res)*sectime_res, sectime_res)
    from scipy.interpolate import interp1d
    f = interp1d(sectime, x, fill_value='extrapolate')
    x_time_ticks = f(time_ticks)
    f = interp1d(sectime, y, fill_value='extrapolate')
    y_time_ticks = f(time_ticks)
    panel.major_ax.plot(x_time_ticks, y_time_ticks, '.', markersize=4, color='k')
    for ind, time_tick in enumerate(time_ticks):
        time = dt0 + datetime.timedelta(seconds=time_tick)
        x_time_tick = x_time_ticks[ind]
        y_time_tick = y_time_ticks[ind]

        if x_time_tick < panel._extent[0] or x_time_tick > panel._extent[1]:
            continue
        if y_time_tick < panel._extent[2] or y_time_tick > panel._extent[3]:
            continue

        panel.major_ax.text( x_time_tick, y_time_tick, time.strftime("%d %H:%M"), fontsize=6)

    panel.add_coastlines()
    panel.add_grids()

    plt.gcf().suptitle('Swarm-C electron density\n 2016-02-03T07:00 - 2016-02-03T07:50')
    plt.savefig('test_n_e.png', dpi=300)
    plt.show()
    
def show_T_e(dt_fr, dt_to):
    import cdflib
    filepath = "~/tmp/SW_OPER_EFIC_LP_1B_20160203T000000_20160203T235959_0501_MDR_EFI_LP.cdf"
    cf = cdflib.CDF(filepath)
    cf_info = cf.cdf_info()
    n_e = cf.varget(variable='Ne')
    T_e = cf.varget(variable='Te')
    sc_lat = cf.varget(variable='Latitude')
    sc_lon = cf.varget(variable='Longitude')
    timestamp = cf.varget(variable='Timestamp')
    
    dtstrs = cdflib.cdfepoch.encode(timestamp)
    dts = np.empty_like(timestamp, dtype=datetime.datetime)
    for ind, dtstr in enumerate(dtstrs):
        dts[ind] = datetime.datetime.strptime(dtstr+'000', '%Y-%m-%dT%H:%M:%S.%f')
    ind_dt = np.where((dts >= dt_fr) & (dts <= dt_to))[0]
    # times = cdflib.cdfepoch.unixtime(timestamp, to_np=True)
    
    
    sc_lon = sc_lon[ind_dt]
    sc_lat = sc_lat[ind_dt]
    sc_dt = dts[ind_dt]
    rho_n_sc = T_e[ind_dt]

    plt.figure(figsize=(8,8))
    panel = PolarView(cs='GEO', pole='N', lon_c=None, lst_c=0, ut=dt_fr, boundary_lat=0., proj_style='Stereographic')
    panel.add_subplot(major=True)
    panel.set_extent(boundary_style='circle')

    data = panel.proj.transform_points(ccrs.PlateCarree(), sc_lon, sc_lat)

    x = data[:, 0]
    y = data[:, 1]


    from scipy.stats import linregress
    coef = linregress(x, y)
    a = coef.slope
    b = coef.intercept

    # x1 = np.linspace(np.nanmin(x), np.nanmax(x), num=500)
    # y1 = np.linspace(np.nanmin(y), np.nanmax(y), num=500)
    # x2d, y2d = np.meshgrid(x, y)
    # z2d = griddata(data[:, 0:2], rho_n_sc.flatten(), (x2d, y2d), method='nearest')
    # z2d_dist = np.abs(a*x2d - y2d + b) / np.sqrt(a**2 + 1)
    # z2d = np.where(z2d_dist>1000, np.nan, z2d)
    #
    # im = panel.major_ax.pcolormesh(x2d, y2d, z2d, vmin=2e-14, vmax=20e-13, cmap='gist_ncar')

    #  panel.major_ax.plot(sc_lon, sc_lat, transform=ccrs.Geodetic(), linewidth=0.5, color='k')
    # plt.colorbar(im)

    # xx = np.tile(x, [3, 1])
    #
    # yy = np.concatenate((y[np.newaxis, :]-150000, y[np.newaxis, :], y[np.newaxis, :]+150000), axis=0)
    # zz = np.concatenate((rho_n_sc.T, rho_n_sc.T, rho_n_sc.T), axis=0)
    #
    # zz = griddata(data[:, 0:2], rho_n_sc.flatten(), (xx, yy), method='nearest')
    # im = panel.major_ax.pcolormesh(xx, yy, zz, vmin=2e-14, vmax=20e-13, cmap='gist_ncar')
    #
    # #  panel.major_ax.plot(sc_lon, sc_lat, transform=ccrs.Geodetic(), linewidth=0.5, color='k')
    # plt.colorbar(im)

    #panel.major_ax.plot(data[:,0], data[:,1], linewidth=3)

    from matplotlib.collections import LineCollection
    from matplotlib.colors import ListedColormap, BoundaryNorm

    data = panel.proj.transform_points(ccrs.PlateCarree(), sc_lon, sc_lat)
    x = data[:, 0]
    y = data[:, 1]
    z = rho_n_sc.flatten()
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    norm = plt.Normalize(500, 4000)
    # norm = colors.LogNorm(vmin=5e2, vmax=1e6)
    lc = LineCollection(segments, cmap='gist_ncar', norm=norm)
    lc.set_array(z)
    lc.set_linewidth(6)
    line = panel.major_ax.add_collection(lc)
    cbar = plt.gcf().colorbar(line, ax=panel.major_ax, pad=0.1, fraction=0.03)
    cbar.set_label(r'$T_e$' + '\n' + r'(K)', rotation=270, labelpad=25)
    #cbaxes = plt.gcf().add_axes([0.8, 0.1, 0.03, 0.8]) 
    #cb = plt.colorbar(panel.major_ax, cax = cbaxes)  


    sectime, dt0 = du.convert_datetime_to_sectime(sc_dt, datetime.datetime(dt_fr.year, dt_fr.month, dt_fr.day))
    sectime_res = 10 * 60
    time_ticks = np.arange(np.floor(sectime[0]/sectime_res)*sectime_res, np.ceil(sectime[-1]/sectime_res)*sectime_res, sectime_res)
    from scipy.interpolate import interp1d
    f = interp1d(sectime, x, fill_value='extrapolate')
    x_time_ticks = f(time_ticks)
    f = interp1d(sectime, y, fill_value='extrapolate')
    y_time_ticks = f(time_ticks)
    panel.major_ax.plot(x_time_ticks, y_time_ticks, '.', markersize=4, color='k')
    for ind, time_tick in enumerate(time_ticks):
        time = dt0 + datetime.timedelta(seconds=time_tick)
        x_time_tick = x_time_ticks[ind]
        y_time_tick = y_time_ticks[ind]

        if x_time_tick < panel._extent[0] or x_time_tick > panel._extent[1]:
            continue
        if y_time_tick < panel._extent[2] or y_time_tick > panel._extent[3]:
            continue

        panel.major_ax.text( x_time_tick, y_time_tick, time.strftime("%d %H:%M"), fontsize=6)

    panel.add_coastlines()
    panel.add_grids()

    plt.gcf().suptitle('Swarm-C electron temperature\n 2016-02-03T07:00 - 2016-02-03T07:50')
    plt.savefig('test_T_e.png', dpi=300)
    plt.show()

if __name__ == "__main__":
    dt_fr = datetime.datetime(2016, 2, 3, 7)
    dt_to = datetime.datetime(2016, 2, 3, 7, 50)

    # show_rho_n(dt_fr, dt_to)
    
    # show_n_e(dt_fr, dt_to)
    
    show_T_e(dt_fr, dt_to)


import numpy as np
from geospacelab.observatory.constants import *


def calc_great_circle_distance(glat_1, glon_1, glat_2, glon_2, r=R_E, formula='Vincenty', unit='degree'):
    factor = np.pi / 180.

    if unit == 'degree':
        glat_1 = glat_1 * factor
        glon_1 = glon_1 * factor
        glat_2 = glat_2 * factor
        glon_2 = glon_2 * factor
    delta_glon = glon_2 - glon_1

    # for a distance more than several kilometers
    if formula == 'classic':
        theta = np.arccos(
            np.sin(glat_1)*np.sin(glat_2) +
            np.cos(glat_1)*np.cos(glat_2)*np.cos(np.abs(delta_glon))
            )
    elif formula == 'Vincenty':
        theta = np.arctan(np.sqrt((np.cos(glat_2)*np.sin(delta_glon))**2
                                  + ((np.cos(glat_1)*np.sin(glat_2)
                                      - np.sin(glat_1)*np.cos(glat_2)*np.cos(delta_glon))**2))
                          / (np.sin(glat_1)*np.sin(glat_2) + np.cos(glat_1)*np.cos(glat_2)*np.cos(delta_glon)))
    else:
        raise NotImplementedError

    d = r * theta
    return d



import datetime
import matplotlib.pyplot as plt

import geospacelab.visualization.eiscat_viewer as eiscat


def test_UHF_CP2():
    dt_fr = datetime.datetime.strptime('20210304' + '2300', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20210305' + '0300', '%Y%m%d%H%M')

    site = 'UHF'
    antenna = 'UHF'
    modulation = 'ant'
    load_mode = 'AUTO'
    data_file_type = 'eiscat-hdf5'

    viewer = eiscat.EISCATViewer(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)

    # select beams before assign the variables
    # viewer.dataset.select_beams(field_aligned=True)
    viewer.dataset.select_beams(az_el_pairs=[(188.6, 77.7)])

    n_e = viewer.assign_variable('n_e')
    T_i = viewer.assign_variable('T_i')
    T_e = viewer.assign_variable('T_e')
    v_i = viewer.assign_variable('v_i_los')
    az = viewer.assign_variable('az')
    el = viewer.assign_variable('el')

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    viewer.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3])
    viewer.draw()
    viewer.add_title()
    viewer.add_panel_labels()
    return viewer


def test_esr_32m():
    dt_fr = datetime.datetime.strptime('20210304' + '2300', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20210305' + '0300', '%Y%m%d%H%M')

    site = 'ESR'
    antenna = '42m'
    modulation = 'ant'
    load_mode = 'AUTO'
    data_file_type = 'eiscat-hdf5'

    viewer = eiscat.EISCATViewer(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)

    # select beams before assign the variables
    # viewer.dataset.select_beams(field_aligned=False)

    n_e = viewer.assign_variable('n_e')
    T_i = viewer.assign_variable('T_i')
    T_e = viewer.assign_variable('T_e')
    v_i = viewer.assign_variable('v_i_los')
    az = viewer.assign_variable('az')
    el = viewer.assign_variable('el')

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    viewer.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3])
    viewer.draw()
    viewer.add_title()
    viewer.add_panel_labels()
    return viewer


def test_uhf_cp3():
    dt_fr = datetime.datetime.strptime('20210322' + '2000', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20210322' + '2300', '%Y%m%d%H%M')

    site = 'UHF'
    antenna = 'UHF'
    modulation = 'ant'
    load_mode = 'AUTO'
    data_file_type = 'eiscat-hdf5'

    viewer = eiscat.EISCATViewer(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)

    # select beams before assign the variables
    # viewer.dataset.select_beams(field_aligned=False)

    n_e = viewer.assign_variable('n_e')
    n_e.visual.axis[0].data_res = 180 # time resolution in seconds
    n_e.visual.axis[1].data = '@d.GEO_LAT.value'
    n_e.visual.axis[1].label = 'GLAT'
    n_e.visual.axis[1].unit = 'deg'
    n_e.visual.axis[1].lim = [65, 75]
    T_i = viewer.assign_variable('T_i')
    T_i.visual.axis[0].data_res = 180
    T_e = viewer.assign_variable('T_e')
    T_e.visual.axis[0].data_res = 180

    v_i = viewer.assign_variable('v_i_los')
    v_i.visual.axis[0].data_res = 180
    v_i.visual.axis[1].data = '@d.GEO_LAT.value'
    v_i.visual.axis[1].label = 'GLAT'
    v_i.visual.axis[1].unit = 'deg'
    v_i.visual.axis[1].lim = [65, 75]

    az = viewer.assign_variable('az')
    az.visual.axis[0].data_res = 180
    el = viewer.assign_variable('el')
    el.visual.axis[0].data_res = 180

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    viewer.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3])
    viewer.draw()
    viewer.add_title()
    viewer.add_panel_labels()
    return viewer


def test_vhf_lowel():
    dt_fr = datetime.datetime.strptime('20060903' + '2030', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20060903' + '2359', '%Y%m%d%H%M')

    site = 'VHF'
    antenna = 'VHF'
    modulation = '60'
    load_mode = 'AUTO'
    data_file_type = 'eiscat-hdf5'

    viewer = eiscat.EISCATViewer(dt_fr, dt_to, site=site, antenna=antenna, modulation=modulation,
                                 data_file_type=data_file_type, load_mode=load_mode)

    # select beams before assign the variables
    # viewer.dataset.select_beams(field_aligned=False)

    n_e = viewer.assign_variable('n_e')
    n_e.visual.axis[0].data_res = 180 # time resolution in seconds
    n_e.visual.axis[1].data = '@d.AACGM_LAT.value'
    n_e.visual.axis[1].label = 'MALT'
    n_e.visual.axis[1].unit = 'deg'
    n_e.visual.axis[1].lim = [68, 74.5]
    T_i = viewer.assign_variable('T_i')
    T_i.visual.axis[0].data_res = 180 # time resolution in seconds
    T_i.visual.axis[1].data = '@d.GEO_LAT.value'
    T_i.visual.axis[1].label = 'GLAT'
    T_i.visual.axis[1].unit = 'deg'
    T_i.visual.axis[1].lim = [71, 75]
    T_e = viewer.assign_variable('T_e')
    T_e.visual.axis[0].data_res = 180 # time resolution in seconds
    T_e.visual.axis[1].data = '@d.GEO_LAT.value'
    T_e.visual.axis[1].label = 'GLAT'
    T_e.visual.axis[1].unit = 'deg'
    T_e.visual.axis[1].lim = [71, 75]
    v_i = viewer.assign_variable('v_i_los')
    v_i.visual.axis[0].data_res = 180 # time resolution in seconds
    v_i.visual.axis[1].data = '@d.GEO_LAT.value'
    v_i.visual.axis[1].label = 'GLAT'
    v_i.visual.axis[1].unit = 'deg'
    v_i.visual.axis[1].lim = [None, None]
    az = viewer.assign_variable('az')
    el = viewer.assign_variable('el')

    layout = [[n_e], [T_e], [T_i], [v_i], [az, el]]
    viewer.set_layout(panel_layouts=layout, row_height_scales=[5, 5, 5, 5, 3])
    viewer.draw()
    viewer.add_title()
    viewer.add_panel_labels()
    return viewer

if __name__ == "__main__":
    test_vhf_lowel()
    test_uhf_cp3()
    test_UHF_CP2()
    test_esr_32m()
    plt.show()


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np
import datetime
import cartopy.crs as ccrs
import re
import copy
from cartopy.mpl.ticker import (
    LongitudeLocator, LatitudeLocator,
    LongitudeFormatter, LatitudeFormatter)
import matplotlib.ticker as mticker
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from scipy.interpolate import interp1d, griddata
import matplotlib.cm as cm
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER

import geospacelab.visualization.mpl as mpl
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.numpymath as mathtool
import geospacelab.toolbox.utilities.pybasic as pybasic
from geospacelab.visualization.mpl.geomap.__base__ import GeoPanelBase
from geospacelab.datahub.__variable_base__ import VariableBase


class GeoPanel(GeoPanelBase):

    def __init__(self, *args, proj_class=None, proj_config: dict = None, **kwargs):

        super().__init__(*args, proj_class=proj_class, proj_config=proj_config, **kwargs)


class PolarMapPanel(GeoPanel):

    def __init__(self, *args, cs='AACGM', style=None, lon_c=None, pole='N', ut=None, lst_c=None, mlt_c=None,
                 mlon_c=None,
                 boundary_lat=30., boundary_style='circle',
                 mirror_south=False,
                 proj_type='Stereographic', **kwargs):
        if style is None:
            style = input("Specify the mapping style: lon-fixed, lst-fixed, mlon-fixed, or mlt-fixed? ")
        self.style = style
        if style in ['lon-fixed']:
            if lon_c is None:
                raise ValueError
            lst_c = None
            mlon_c = None
            mlt_c = None

        if style in ['lst-fixed']:
            if lst_c is None:
                raise ValueError
            lon_c = None
            mlon_c = None
            mlt_c = None

        if style in ['mlon-fixed']:
            if mlon_c is None:
                raise ValueError
            lst_c = None
            lon_c = mlon_c
            mlt_c = None

        if style in ['mlt-fixed']:
            if mlt_c is None:
                raise ValueError
            lst_c = None
            lon_c = None
            mlon_c = None

        # if lon_c is not None and pole == 'S':
        #    lon_c = lon_c + 180.

        self.lat_c = None
        self.lon_c = lon_c
        self.ut = ut
        self.boundary_lat = boundary_lat
        self.boundary_style = boundary_style
        self.pole = pole
        self.lst_c = lst_c
        self.cs = cs
        self.depend_mlt = False
        self.mlt_c = mlt_c
        self.mirror_south = mirror_south
        self._extent = None
        proj_class = getattr(ccrs, proj_type)
        proj_config = {'central_latitude': self.lat_c, 'central_longitude': self.lon_c}
        super().__init__(*args, proj_class=proj_class, proj_config=proj_config, **kwargs)
        self.set_extent()

    @staticmethod
    def _transform_mlt_to_lon(mlt):
        lon = mlt / 24. * 360.
        lon = np.mod(lon, 360.)
        return lon
    
    @staticmethod
    def _transform_lst_to_lon(lst):
        lon = lst / 24. * 360.
        lon = np.mod(lon, 360.)
        return lon

    def add_axes(self, *args, major=False, label=None, **kwargs):
        if major:
            kwargs.setdefault('projection', self.projection)
        ax = super().add_axes(*args, major=major, label=label, **kwargs)
        return ax

    # def add_lands(self):
    #     import cartopy.io.shapereader as shpreader
    #
    #     resolution = '110m'
    #     shpfilename = shpreader.natural_earth(resolution=resolution,
    #                                           category='physical',
    #                                           name='coastline')
    #     reader = shpreader.Reader(shpfilename)
    #     lands = list(reader.geometries())
    #
    #     for ind1, land in enumerate(lands):
    #         land_polygons = list(land)
    #         for ind2, polygon in enumerate(land_polygons):
    #             x, y = polygon.exterior.coords.xy
    #
    def cs_transform(self, cs_fr=None, cs_to=None, coords=None, ut=None):
        import geospacelab.cs as geo_cs

        if cs_to is None:
            cs_to = self.cs
        if ut is None:
            ut = self.ut
        cs_class = getattr(geo_cs, cs_fr.upper())
        cs1 = cs_class(coords=coords, ut=ut)
        if cs_fr != cs_to:
            cs2 = cs1(cs_to=cs_to, append_mlt=self.depend_mlt)
        else:
            cs2 = cs1
        if self.style == 'lst-fixed':
            try:
                cs2.coords.lon = self._transform_lst_to_lon(cs2.coords.lst)
            except:
                mylog.StreamLogger.warning("LST is not specified!")
        if self.depend_mlt:
            cs2.coords.lon = self._transform_mlt_to_lon(cs2.coords.mlt)
        return cs2

    def overlay_coastlines(self, linestyle='-', linewidth=0.5, color='#797A7D', zorder=100, alpha=0.7, **kwargs):
        import cartopy.io.shapereader as shpreader

        resolution = '110m'
        shpfilename = shpreader.natural_earth(resolution=resolution,
                                              category='physical',
                                              name='coastline')
        coastlines = list(shpreader.Reader(shpfilename).geometries())

        x = np.array([])
        y = np.array([])
        for ind, c in enumerate(coastlines[:-1]):
            # print(ind)
            # print(len(c.xy[0]))
            # if ind not in [4013, 4014]:
            #    continue
            x0 = np.array(c.xy[0])
            y0 = np.array(c.xy[1])
            # if len(x0) < 20:  # omit small islands, etc.
            #    continue
            x0 = np.mod(x0[::1], 360)
            y0 = y0[::1]
            x = np.append(np.append(x, x0), np.nan)
            y = np.append(np.append(y, y0), np.nan)

            # csObj = scs.SpaceCS(x, y, CS='GEO', dt=self.dt, coords_labels=['lon', 'lat'])

        coords = {'lat': y, 'lon': x, 'height': np.ones_like(y) * 0}
        coords = self.cs_transform(cs_fr='GEO', cs_to=self.cs, coords=coords)
        x_new = coords['lon']
        y_new = coords['lat']
        # x_new, y_new = x, y
        self().plot(
            x_new, y_new,
            transform=ccrs.Geodetic(),
            linestyle=linestyle, linewidth=linewidth, color=color, zorder=zorder, alpha=alpha,
            **kwargs)
        # self.ax.scatter(x_new, y_new, transform=self.default_transform,
        #             marker='.', edgecolors='none', color='#C0C0C0', s=1)

        return

    def overlay_gridlines(self,
                      lat_res=None, lon_res=None,
                      lat_label_separator=None, lon_label_separator=None,
                      lat_label_format=None, lon_label_format=None,
                      lat_label=True, lat_label_clock=6.5, lat_label_config={},
                      lon_label=True, lon_label_config={},
                      gridlines_config={}):

        default_gridlines_label_config = {
            'lon-fixed': {
                'lat_res': 10.,
                'lon_res': 30.,
                'lat_label_separator': 0,
                'lon_label_separator': 0,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%d%D%E',
            },
            'mlon-fixed': {
                'lat_res': 10.,
                'lon_res': 30.,
                'lat_label_separator': 0,
                'lon_label_separator': 0,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%d%D%E',
            },
            'lst-fixed': {
                'lat_res': 10.,
                'lon_res': 15.,
                'lat_label_separator': 0,
                'lon_label_separator': 2,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%02d LT',
            },
            'mlt-fixed': {
                'lat_res': 10.,
                'lon_res': 15.,
                'lat_label_separator': 0,
                'lon_label_separator': 2,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%02d MLT',
            },
        }

        def label_formatter(value, fmt=''):
            ms = re.findall(r"(%[0-9]*[a-zA-Z])", fmt)
            fmt_new = copy.deepcopy(fmt)
            patterns = []
            if "%N" in ms:
                if value > 0:
                    lb_ns = 'N'
                elif value < 0:
                    lb_ns = 'S'
                else:
                    lb_ns = ''
                value = np.abs(value)

            if "%E" in ms:
                mod_value = np.mod(value - 180, 360) - 180.
                if mod_value == 0 or mod_value == 180:
                    lb_ew = ''
                elif mod_value < 0:
                    lb_ew = 'W'
                else:
                    lb_ew = 'E'
                value = np.abs(mod_value)

            for ind_m, m in enumerate(ms):
                if m == "%N":
                    fmt_new = fmt_new.replace(m, '{}')
                    patterns.append(lb_ns)
                elif m == "%E":
                    fmt_new = fmt_new.replace(m, '{}')
                    patterns.append(lb_ew)
                elif m == "%D":
                    continue
                else:
                    if 'd' in m:
                        value = int(value)
                    fmt_new = fmt_new.replace(m, '{:' + m[1:] + '}')
                    patterns.append(value)
            string_new = fmt_new.format(*patterns)
            if "%D" in string_new:
                splits = string_new.split("%D")
                string_new = splits[0] + r'$^{\circ}$' + splits[1]

            return string_new

        default_label_config = default_gridlines_label_config[self.style]
        if lat_res is None:
            lat_res = default_label_config['lat_res']
        if lon_res is None:
            lon_res = default_label_config['lon_res']
        if lat_label_separator is None:
            lat_label_separator = default_label_config['lat_label_separator']
        if lon_label_separator is None:
            lon_label_separator = default_label_config['lon_label_separator']
        if lat_label_format is None:
            lat_label_format = default_label_config['lat_label_format']
        if lon_label_format is None:
            lon_label_format = default_label_config['lon_label_format']

        # xlocator
        if self.style == 'lst-fixed':
            lon_shift = (self.ut - datetime.datetime(
                self.ut.year, self.ut.month, self.ut.day, self.ut.hour
            )).total_seconds() / 86400. * 360
        else:
            lon_shift = 0
        num_lons = 360. // lon_res + 1.
        lons = np.linspace(-180., 180., int(num_lons)) - lon_shift
        xlocator = mticker.FixedLocator(lons)

        # ylocator
        lat_fr = np.abs(self.boundary_lat) + np.mod(90 - np.abs(self.boundary_lat), lat_res)
        lat_to = 90.
        # num_lats = (lat_to - lat_fr) / lat_res + 1.
        lats = np.arange(lat_fr, lat_to, lat_res) * np.sign(self.lat_c)
        ylocator = mticker.FixedLocator(lats)

        pybasic.dict_set_default(gridlines_config, color='#331900', linewidth=0.5, linestyle=':',
                                     draw_labels=False)
        gl = self().gridlines(crs=ccrs.PlateCarree(), **gridlines_config)

        gl.xlocator = xlocator
        gl.ylocator = ylocator

        if lat_label:
            pybasic.dict_set_default(lat_label_config, fontsize=plt.rcParams['font.size'] - 2, fontweight='roman',
                    ha='center', va='center',
                    family='fantasy', alpha=0.9
                )
            if self.pole == 'S':
                clock = lat_label_clock / 12 * 360
                rotation = 180 - clock
                if self.mirror_south:
                    clock = - clock
                    rotation = 180 + clock
            else:
                clock = np.mod(180. - lat_label_clock / 12 * 360, 360)
                rotation = clock
            label_lon = self.lon_c + clock
            for ind, lat in enumerate(lats):
                if np.mod(ind, lat_label_separator+1) != 0:
                    continue
                # if lat > 0:
                #     label = r'' + '{:d}'.format(int(lat)) + r'$^{\circ}$N'
                # elif lat < 0:
                #     label = r'' + '{:d}'.format(int(-lat)) + r'$^{\circ}$S'
                # else:
                #     label = r'' + '{:d}'.format(int(0)) + r'$^{\circ}$S'
                label = label_formatter(lat, fmt=lat_label_format)
                self().text(
                    label_lon, lat, label, transform=ccrs.PlateCarree(),
                    rotation=rotation,
                    **lat_label_config
                )

        if lon_label:
            pybasic.dict_set_default(
                lon_label_config,
                va='center', ha='center',
                family='fantasy', fontweight='ultralight')
            lb_lons = np.arange(0, 360., lon_res)
            if self.pole == 'S':
                lon_shift_south = 180.
            else:
                lon_shift_south = 0
            if self.style == 'lon-fixed' or self.style == 'mlon-fixed':
                lb_lons_loc = lb_lons
            elif self.style == 'lst-fixed':
                lb_lons_loc = np.mod(self.lon_c + (lb_lons - self.lst_c / 24 * 360) + lon_shift_south, 360)
            elif self.style == 'mlt-fixed':
                lb_lons_loc = lb_lons

            if self.boundary_style == 'circle':
                lb_lats = np.empty_like(lb_lons)
                lb_lats[:] = self.boundary_lat
                data = self.projection.transform_points(ccrs.PlateCarree(), lb_lons_loc, lb_lats)
                xdata = data[:, 0]
                ydata = data[:, 1]
                scale = (self._extent[1] - self._extent[0]) * 0.03
                xdata = xdata + scale * np.sin((lb_lons_loc - self.lon_c) * np.pi / 180.)
                ydata = ydata - np.sign(self.lat_c) * scale * np.cos((lb_lons_loc - self.lon_c) * np.pi / 180.)
                for ind, lb_lon in enumerate(lb_lons):
                    if np.mod(ind, lon_label_separator+1) != 0:
                        continue
                    x = xdata[ind]
                    y = ydata[ind]
                    # if self.style in ['lon-fixed', 'mlon-fixed']:
                    #     lb_lon = np.mod(lb_lon + 180, 360) - 180
                    #     if lb_lon == 0 or np.abs(lb_lon) == 180.:
                    #         label = r'' + '{:d}'.format(int(np.abs(lb_lon))) + r'$^{\circ}$'
                    #     if lb_lon < 0:
                    #         label = r'' + '{:d}'.format(int(-lb_lon)) + r'$^{\circ}$W'
                    #     else:
                    #         label = r'' + '{:d}'.format(int(lb_lon)) + r'$^{\circ}$E'
                    # elif self.style == 'mlt-fixed':
                    #     label = '{:d}'.format(int(lb_lon / 15)) + ' MLT'
                    # elif self.style == 'lst-fixed':
                    #     label = '{:d}'.format(int(lb_lon / 15)) + ' LT'
                    if self.style in ['lon-fixed', 'mlon-fixed']:
                        lb_value = lb_lon
                    elif self.style in ['lst-fixed', 'mlt-fixed']:
                        lb_value = lb_lon / 15
                    label = label_formatter(lb_value, fmt=lon_label_format)
                    if self.pole == 'S':
                        rotation = self.lon_c - lb_lons_loc[ind]
                        if self.mirror_south:
                            rotation = -self.lon_c + lb_lons_loc[ind]
                    else:
                        rotation = (lb_lons_loc[ind] - self.lon_c) + 180
                    self().text(
                        x, y, label,
                        rotation=rotation,
                        **lon_label_config,
                    )

        # gl.xformatter = LONGITUDE_FORMATTER()
        # gl.yformatter = LATITUDE_FORMATTER()
        # for ea in gl.label_artists:
        #     if ea[1]==False:
        #         tx = ea[2]
        #         xy = tx.get_position()
        #         #print(xy)
        #
        #         tx.set_position([30, xy[1]])

    def set_extent(self, boundary_lat=None, boundary_style=None):
        if boundary_lat is not None:
            self.boundary_lat = boundary_lat
        if boundary_style is not None:
            self.boundary_style = boundary_style
        x = np.array([270., 90., 180., 0.])
        y = np.ones(4) * self.boundary_lat
        x = np.arange(0., 360., 5.)
        y = np.empty_like(x)
        y[:] = 1. * self.boundary_lat
        data = self.projection.transform_points(ccrs.PlateCarree(), x, y)
        # self.axes.plot(x, y, '.', transform=ccrs.PlateCarree())
        ext = [np.nanmin(data[:, 0]), np.nanmax(data[:, 0]), np.nanmin(data[:, 1]), np.nanmax(data[:, 1])]
        self._extent = ext
        self().set_extent(ext, self.projection)

        self._set_boundary_style()

        self._check_mirror_south()

    def _check_mirror_south(self):
        if self.pole == 'S' and self.mirror_south:
            xlim = self.major_ax.get_xlim()
            self.major_ax.set_xlim([max(xlim), min(xlim)])

    def _set_boundary_style(self):

        style = self.boundary_style
        if style == 'square':
            return
        elif style == 'circle':
            theta = np.linspace(0, 2 * np.pi, 400)
            center = self.projection.transform_point(self.lon_c, self.lat_c, ccrs.PlateCarree())
            radius = (self._extent[1] - self._extent[0]) / 2
            verts = np.vstack([np.sin(theta), np.cos(theta)]).T
            circle = mpath.Path(verts * radius + center)
            self().set_boundary(circle, transform=self.projection)
        else:
            raise NotImplementedError

    def overlay_pcolormesh(
            self, data=None, coords=None, cs=None, *,
            ut = None,
            regridding: bool = True,
            data_res: float = 0.7,
            grid_res: float = 0.1,
            interp_method: str = 'linear',
            c_scale='linear', c_lim=None,
            **kwargs):

        kwargs.setdefault('shading', 'auto')
        if c_lim is None:
            c_lim = [0, 0]
            c_lim[0] = np.nanmin(data.flatten())
            c_lim[1] = np.nanmax(data.flatten())

        if c_scale == 'log':
            norm = mcolors.LogNorm(vmin=c_lim[0], vmax=c_lim[1])
            kwargs.update(norm=norm)
        else:
            kwargs.update(vmin=c_lim[0])
            kwargs.update(vmax=c_lim[1])

        cs_new = self.cs_transform(cs_fr=cs, coords=coords, ut=ut)
        lat_in = cs_new['lat']
        lon_in = cs_new['lon']
        data_in = data
        if regridding:
            lon_in, lat_in, data_in = self.grid_data(lon_in, lat_in, data_in, data_res=data_res, grid_res=grid_res)
            transform = None
            # self.boundary_lat
            #
            # grid_lat_res = kwargs.pop('grid_lat_res', 0.1)
            # grid_lon_res = kwargs.pop('grid_lon_res', 0.1)
            #
            # ind_data = np.where(np.isfinite(data))
            # data_pts = data[ind_data]
            # lat_pts = cs_new['lat'][ind_data]
            # lon_pts = cs_new['lon'][ind_data]
            #
            # grid_lon, grid_lat = np.meshgrid(
            #     np.arange(0., 360., grid_lon_res),
            #     np.append(np.arange(self.boundary_lat, self.lat_c, np.sign(self.lat_c) * grid_lat_res), self.lat_c)
            # )
            #
            # grid_data = griddata((lon_pts, lat_pts), data_pts, (grid_lon, grid_lat), method='nearest')
            # grid_data_lat = griddata(
            #     (lon_pts, lat_pts), lat_pts, (grid_lon, grid_lat), method='nearest'
            # )
            # grid_data_lon = griddata(
            #     (lon_pts, lat_pts), lon_pts, (grid_lon, grid_lat), method='nearest'
            # )
            # factor = np.pi / 180.
            # big_circle_d = 6371. * np.arccos(
            #     np.sin(grid_data_lat * factor) * np.sin(grid_lat * factor) +
            #     np.cos(grid_data_lat * factor) * np.cos(grid_lat * factor) * np.cos(
            #         np.abs((grid_lon - grid_data_lon) * factor))
            # )
            # grid_data = np.where(big_circle_d < data_lat_res*111.2, grid_data, np.nan)
            # lat_in = grid_lat
            # lon_in = grid_lon
            # data_in = grid_data
        else:
            transform = ccrs.PlateCarree()

        ipm = self().pcolormesh(lon_in, lat_in, data_in, transform=transform, **kwargs)

        return ipm

    def overlay_contour(
            self,
            data, coords=None, cs=None,
            regridding=True, data_res=1, grid_res=0.1, interp_method='linear',
            **kwargs,
    ):
        levels = kwargs.pop('levels', 10)
        if cs is None:
            cs = self.cs

        cs_new = self.cs_transform(cs_fr=cs, coords=coords)
        lat_in = cs_new['lat']
        lon_in = cs_new['lon']
        data_in = data
        if regridding:
            lon_in, lat_in, data_in = self.grid_data(lon_in, lat_in, data_in, data_res=data_res, grid_res=grid_res, interp_method=interp_method)
            transform = None
        else:
            transform = ccrs.PlateCarree()

        ict = self.major_ax.contour(lon_in, lat_in, data_in, levels, transform=transform, **kwargs)
        return ict

    def grid_data(self, lon_in, lat_in, data_in, grid_res=0.1, data_res=0.5, interp_method='linear'):
        if self.pole == "N":
            sign_lat = 1.
        else:
            sign_lat = -1.
        ind_data = np.where(np.isfinite(data_in) & (lat_in * sign_lat > (self.boundary_lat-5.)) & np.isfinite(lon_in))
        data_pts = data_in[ind_data]
        lat_pts = lat_in[ind_data]
        lon_pts = lon_in[ind_data]

        pos_data = self.projection.transform_points(ccrs.PlateCarree(), lon_pts, lat_pts)
        pos_x = np.float32(pos_data[:, 0])
        pos_y = np.float32(pos_data[:, 1])

        lim=[]
        xlim = list(self.major_ax.get_xlim())
        ylim = list(self.major_ax.get_ylim())
        lim.extend(xlim)
        lim.extend(ylim)
        lim = [np.float32(x) for x in lim]
        width = np.max(lim) - np.min(lim)
        lat_width = (90. - np.abs(self.boundary_lat)) * 2
        grid_x_res = grid_res / lat_width * width
        if xlim[0] > xlim[1]:
            grid_x_res = -grid_x_res
        grid_x, grid_y = np.meshgrid(
            np.arange(lim[0], lim[1], grid_x_res),
            np.arange(lim[0], lim[1], grid_x_res)
        )

        grid_data = griddata((pos_x, pos_y), data_pts, (grid_x, grid_y), method=interp_method)
        grid_data_x = griddata(
            (pos_x, pos_y), pos_x, (grid_x, grid_y), method='nearest'
        )
        grid_data_y = griddata(
            (pos_x, pos_y), pos_y, (grid_x, grid_y), method='nearest'
        )

        ind_out = np.where(
            np.sqrt((grid_data_x - grid_x) ** 2 + (grid_data_y - grid_y) ** 2) > data_res / lat_width * width
        )
        grid_data[ind_out] = np.nan

        return grid_x, grid_y, grid_data

    def overlay_sc_trajectory(self, sc_ut=None, sc_coords=None, cs=None, *, show_trajectory=True, color='#EEEEEE',
                          time_tick=True, time_tick_res=300., time_tick_scale=0.05,
                          time_tick_label=True, time_tick_label_format="%H:%M", time_tick_label_fontsize=8,
                          time_tick_label_rotation=45., time_tick_label_offset=0.05,
                          time_minor_tick=True, time_minor_tick_res=60,
                          time_tick_width=1, **kwargs):
        kwargs.setdefault('trajectory_config', {
            'linewidth': 1,
            'linestyle': '-',
            'color': 'k',
            'zorder': max([_.zorder for _ in self.major_ax.get_children()])
        })
        kwargs['trajectory_config'].update(color=color)

        cs_new = self.cs_transform(cs_fr=cs, coords=sc_coords, ut=sc_ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]
        dts_in = sc_ut[ind_lat]

        if self.ut is None:
            self.ut = sc_ut[0]
        if show_trajectory:
            self.major_ax.plot(lon_in, lat_in, transform=ccrs.Geodetic(), **kwargs['trajectory_config'])

        if time_tick:
            data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
            xdata = data[:, 0]
            ydata = data[:, 1]

            sectime, dt0 = dttool.convert_datetime_to_sectime(
                dts_in, datetime.datetime(self.ut.year, self.ut.month, self.ut.day)
            )

            time_ticks = np.arange(np.ceil(sectime[0] / time_tick_res) * time_tick_res,
                                   np.ceil(sectime[-1] / time_tick_res) * time_tick_res, time_tick_res)

            f = interp1d(sectime, xdata, fill_value='extrapolate')
            x_i = f(time_ticks)
            f = interp1d(sectime, ydata, fill_value='extrapolate')
            y_i = f(time_ticks)

            slope = mathtool.calc_curve_tangent_slope(xdata, ydata)
            f = interp1d(sectime, slope, fill_value='extrapolate')
            slope_i = f(time_ticks)
            l = np.empty_like(x_i)
            l[:] = (self._extent[1] - self._extent[0]) * time_tick_scale

            xq = x_i
            yq = y_i
            uq1 = - l * np.sin(slope_i)
            vq1 = l * np.cos(slope_i)

            zorder = kwargs['trajectory_config']['zorder']
            self.major_ax.quiver(
                xq, yq, uq1, vq1,
                units='xy', angles='xy', scale=1., scale_units='xy',
                width=time_tick_width*0.003 * (self._extent[1] - self._extent[0]),
                headlength=0, headaxislength=0, pivot='middle', color=color,
                zorder=zorder
            )

            if time_tick_label:
                offset = time_tick_label_offset * (self._extent[1] - self._extent[0])
                for ind, time_tick in enumerate(time_ticks):
                    time = dt0 + datetime.timedelta(seconds=time_tick)
                    x_time_tick = x_i[ind] - offset * np.sin(slope_i[ind])
                    y_time_tick = y_i[ind] + offset * np.cos(slope_i[ind])

                    self.major_ax.text(
                        x_time_tick, y_time_tick, time.strftime(time_tick_label_format),
                        fontsize=time_tick_label_fontsize,
                        rotation=slope[ind] * 180. / np.pi + time_tick_label_rotation,
                        ha='center', va='center', color=color,
                        zorder=zorder
                    )

            # self.major_ax.plot(x_time_ticks, y_time_ticks, **kwargs['time_tick_config'])

            if time_minor_tick:

                time_ticks = np.arange(np.ceil(sectime[0] / time_minor_tick_res) * time_minor_tick_res,
                                       np.ceil(sectime[-1] / time_minor_tick_res) * time_minor_tick_res, time_minor_tick_res)

                f = interp1d(sectime, xdata, fill_value='extrapolate')
                x_i = f(time_ticks)
                f = interp1d(sectime, ydata, fill_value='extrapolate')
                y_i = f(time_ticks)

                f = interp1d(sectime, slope, fill_value='extrapolate')
                slope_i = f(time_ticks)
                l = np.empty_like(x_i)
                l[:] = (self._extent[1] - self._extent[0]) * time_tick_scale / 2.5

                xq = x_i
                yq = y_i
                uq1 = - l * np.sin(slope_i)
                vq1 = l * np.cos(slope_i)

                self.major_ax.quiver(
                    xq, yq, uq1, vq1,
                    units='xy', angles='xy', scale=1., scale_units='xy',
                    width=time_tick_width*0.002*(self._extent[1] - self._extent[0]),
                    headlength=0, headaxislength=0, pivot='middle', color=color,
                    zorder=zorder
                )

    def overlay_cross_track_vector(
            self, vector, unit_vector, sc_ut=None, sc_coords=None, cs=None, *,
            unit_vector_scale=0.1, vector_unit='',
            color='r', alpha=0.8, quiverkey_config={},
            vector_width=0.5,
            edge='off',
            edge_color=None,
            edge_linestyle='-',
            edge_linewidth=1.,
            edge_marker=None,
            edge_markersize=5,
            edge_alpha=0.8,
            legend='on',
            legend_pos_x=0.9,
            legend_pos_y=0.95,
            legend_label=None,
            legend_linewidth=None,
            **kwargs):

        if edge_color is None:
            edge_color = color

        kwargs.setdefault(
            'zorder', max([_.zorder for _ in self.major_ax.get_children()])
        )

        cs_new = self.cs_transform(cs_fr=cs, coords=sc_coords, ut=sc_ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]
        dts_in = sc_ut[ind_lat]
        vector = vector[ind_lat]

        data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
        xdata = np.float32(data[:, 0])
        ydata = np.float32(data[:, 1])
        slope = mathtool.calc_curve_tangent_slope(xdata, ydata, degree=3)

        width = np.float32(self._extent[1] - self._extent[0])
        vector = vector / unit_vector * width * unit_vector_scale
        sign = np.sign(xdata[1] - xdata[0])
        xq = xdata
        yq = ydata
        uq1 = - sign * vector * np.sin(slope)
        vq1 = sign * vector * np.cos(slope)

        iq = self.major_ax.quiver(
            xq, yq, uq1, vq1,
            units='xy', angles='xy', scale=1., scale_units='xy',
            width=vector_width*0.002 * (self._extent[1] - self._extent[0]),
            headlength=0, headaxislength=0, pivot='tail', color=color, alpha=alpha, **kwargs
        )

        # Add quiverkey
        if legend in ['on', True]:
            if legend_linewidth is None:
                legend_linewidth = vector_width
            if legend_label is None:
                legend_label = str(unit_vector) + ' ' + vector_unit
            quiverkey_config = pybasic.dict_set_default(
                quiverkey_config, X=legend_pos_x, Y=legend_pos_y, U=width*unit_vector_scale,
                width=legend_linewidth * 0.002 * (self._extent[1] - self._extent[0]),
                label=legend_label, color=color,
            )
            X = quiverkey_config.pop('X')
            Y = quiverkey_config.pop('Y')
            U = quiverkey_config.pop('U')
            label = quiverkey_config.pop('label')
            iqk = self.major_ax.quiverkey(
                iq, X, Y, U, label, **kwargs
            )

        if edge == 'on':
            xx = iq.X + iq.U
            yy = iq.Y + iq.V
            self.major_ax.plot(xx, yy, linestyle=edge_linestyle, linewidth=edge_linewidth,
                               color=edge_color, marker=edge_marker, markersize=edge_markersize,
                               alpha=edge_alpha)

        return iq

    def overlay_sc_coloured_line(
            self, z_data, sc_coords=None, sc_ut=None, cs=None, *, c_map='jet',
            c_scale='linear', c_lim=None, line_width=6., **kwargs):
        from matplotlib.collections import LineCollection
        import matplotlib.colors as mpl_color

        kwargs.setdefault(
            'zorder', max([_.zorder for _ in self.major_ax.get_children()])
        )
        if c_lim is None:
            c_lim = [0, 0]
            c_lim[0] = np.nanmin(z_data.flatten())
            c_lim[1] = np.nanmax(z_data.flatten())

        if c_scale == 'log':
            norm = mcolors.LogNorm(vmin=c_lim[0], vmax=c_lim[1])
        else:
            norm = mcolors.Normalize(vmin=c_lim[0], vmax=c_lim[1])

        cs_new = self.cs_transform(cs_fr=cs, coords=sc_coords, ut=sc_ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]
        dts_in = sc_ut[ind_lat]

        pos_data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
        x = np.float32(pos_data[:, 0])
        y = np.float32(pos_data[:, 1])
        z = z_data[ind_lat]

        points = np.array([x, y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)

        lc = LineCollection(segments, norm=norm, cmap=c_map, **kwargs)
        lc.set_array(z)
        lc.set_linewidth(line_width)
        self().add_collection(lc)
        return lc
        # clabel = label + ' (' + unit + ')'
        # self.add_colorbar(self.major_ax, line, cscale=scale, clabel=clabel)
        # cbar = plt.gcf().colorbar(line, ax=panel.major_ax, pad=0.1, fraction=0.03)
        # cbar.set_label(r'$n_e$' + '\n' + r'(cm$^{-3}$)', rotation=270, labelpad=25)

    def overlay_sites(self, site_ids=None, coords=None, cs=None, **kwargs):
        kwargs = pybasic.dict_set_default(kwargs, color='k', linestyle='', markersize=5, marker='.')

        cs_new = self.cs_transform(cs_fr=cs, coords=coords)
        
        isc = self().plot(cs_new['lon'], cs_new['lat'], transform=ccrs.PlateCarree(), **kwargs)
        return isc

    def add_colorbar(self, im, ax=None, figure=None,
                     c_scale='linear', c_label=None,
                     c_ticks=None, c_tick_labels=None, c_tick_label_step=1,
                     left=1.1, bottom=0.1, width=0.05, height=0.7, **kwargs
                     ):
        if figure is None:
            figure = plt.gcf()
        if ax is None:
            ax = self()
        pos = ax.get_position()
        ax_width = pos.x1 - pos.x0
        ax_height = pos.y1 - pos.y0
        ca_left = pos.x0 + ax_width * left
        ca_bottom = pos.y0 + ax_height * bottom
        ca_width = ax_width * width
        ca_height = ax_height * height
        cax = self.add_axes([ca_left, ca_bottom, ca_width, ca_height], major=False, label=c_label)
        cb = figure.colorbar(im, cax=cax, **kwargs)
        ylim = cax.get_ylim()

        cb.set_label(c_label, rotation=270, va='bottom', size='medium')
        if c_ticks is not None:
            cb.ax.yaxis.set_ticks(c_ticks)
            if c_tick_labels is not None:
                cb.ax.yaxis.set_ticklabels(c_tick_labels)
        else:
            if c_scale == 'log':
                num_major_ticks = int(np.ceil(np.diff(np.log10(ylim)))) * 2
                cax.yaxis.set_major_locator(mticker.LogLocator(base=10.0, numticks=num_major_ticks))
                n = c_tick_label_step
                [l.set_visible(False) for (i, l) in enumerate(cax.yaxis.get_ticklabels()) if i % n != 0]
                # [l.set_ha('right') for (i, l) in enumerate(cax.yaxis.get_ticklabels()) if i % n != 0]
                minorlocator = mticker.LogLocator(base=10.0, subs=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9),
                                                     numticks=12)
                cax.yaxis.set_minor_locator(minorlocator)
                cax.yaxis.set_minor_formatter(mticker.NullFormatter())
        cax.yaxis.set_tick_params(labelsize='x-small')
        return [cax, cb]

    @property
    def pole(self):
        return self._pole

    @pole.setter
    def pole(self, value):
        if value.upper() in ['N', 'NORTH', 'NORTHERN']:
            self._pole = 'N'
            self.lat_c = 90.
            self.boundary_lat = np.abs(self.boundary_lat)
        elif value.upper() in ['S', 'SOUTH', 'SOUTHERN']:
            self._pole = 'S'
            self.lat_c = -90.
            self.boundary_lat = - np.abs(self.boundary_lat)
        else:
            raise ValueError

    @property
    def lst_c(self):
        return self._lst_c

    @lst_c.setter
    def lst_c(self, lst):
        self._lst_c = lst
        if lst is None:
            return
        if self.pole == 'N':
            self.lon_c = np.mod((self.lst_c - (self.ut.hour + self.ut.minute / 60)) * 15., 360.)
        elif self.pole == 'S':
            self.lon_c = np.mod((self.lst_c - (self.ut.hour + self.ut.minute / 60)) * 15. + 180., 360.)

    @property
    def mlt_c(self):
        return self._mlt_c

    @mlt_c.setter
    def mlt_c(self, mlt):
        if mlt is not None:
            self.depend_mlt = True
            self.lon_c = self._transform_mlt_to_lon(mlt)
            if self.pole == 'S':
                self.lon_c = np.mod(self.lon_c + 180., 360)
            if self.cs == "GEO":
                raise AttributeError('A magnetic coordinate system must be specified (Set the attribute "cs")!')
        self._mlt_c = mlt



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import geospacelab.toolbox.utilities.pyclass as pyclass


class StrAttrBase(str):
    def __new__(cls, str_in):
        if issubclass(str_in.__class__, StrAttrBase):
            obj = str_in
        else:
            obj = str.__new__(cls, str_in)
        return obj

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)


class DatabaseModel(StrAttrBase):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in)
        obj.category = kwargs.pop('category', '')
        obj.url = kwargs.pop('url', None)
        obj.description = kwargs.pop('description', '')
        obj.notes = kwargs.pop('notes', '')
        return obj


class FacilityModel(StrAttrBase):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in)
        obj.category = kwargs.pop('category', '')
        obj.url = kwargs.pop('url', None)
        obj.description = kwargs.pop('description', '')
        obj.notes = kwargs.pop('notes', '')
        return obj


class ProductModel(StrAttrBase):

    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in)
        obj.category = kwargs.pop('category', '')
        obj.url = kwargs.pop('url', None)
        obj.description = kwargs.pop('description', '')
        obj.notes = kwargs.pop('notes', '')
        return obj


class InstrumentModel(StrAttrBase):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in)
        obj.category = kwargs.pop('category', '')
        obj.description = kwargs.pop('description', '')
        obj.notes = kwargs.pop('notes', '')
        return obj


class SiteModel(StrAttrBase):

    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in)
        obj.category = kwargs.pop('category', '')
        obj.name = kwargs.pop('name', '')
        obj.code = str_in
        obj.location = kwargs.pop('location', {'GEO_LAT': None, 'GEO_LON': None, 'GEO_ALT': None})
        obj.description = kwargs.pop('description', '')
        obj.notes = kwargs.pop('notes', '')
        return obj


class MetadataModel(object):
    def __init__(self):
        self.database = ''

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        self._database = value

    def config(self, logging=True, **kwargs):
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)